#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2007 John Beard john.j.beard@gmail.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""
This extension allows you to draw a polar grid in Inkscape.
There is a wide range of options including subdivision and labels.
"""

from math import cos, log, pi, sin

import inkex
from inkex import Group, Circle, TextElement


def draw_circle(r, cx, cy, width, fill, name, parent):
    """Draw an SVG circle"""
    circle = parent.add(Circle(cx=str(cx), cy=str(cy), r=str(r)))
    circle.style = {"stroke": "#000000", "stroke-width": str(width), "fill": fill}
    circle.label = name


def draw_line(x1, y1, x2, y2, width, name, parent):
    """Draw an SVG line"""
    line = parent.add(inkex.PathElement())
    line.style = {"stroke": "#000000", "stroke-width": str(width), "fill": "none"}
    line.path = "M {},{} L {},{}".format(x1, y1, x2, y2)
    line.label = name


def draw_label(x, y, string, font_size, name, parent):
    """Draw a centered label"""
    label = parent.add(TextElement(x=str(x), y=str(y)))
    label.style = {
        "text-align": "center",
        "vertical-align": "top",
        "text-anchor": "middle",
        "font-size": str(font_size) + "px",
        "fill-opacity": "1.0",
        "stroke": "none",
        "font-weight": "normal",
        "font-style": "normal",
        "fill": "#000000",
    }
    label.text = string
    label.label = name


class GridPolar(inkex.GenerateExtension):
    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument("--r_divs", type=int, default=5, help="Circular Divisions")
        pars.add_argument(
            "--dr", type=float, default=50, help="Circular Division Spacing"
        )
        pars.add_argument(
            "--r_subdivs", type=int, default=3, help="Subdivisions per Major div"
        )
        pars.add_argument(
            "--r_log", type=inkex.Boolean, default=False, help="Logarithmic subdiv"
        )
        pars.add_argument(
            "--r_divs_th", type=float, default=2, help="Major Line thickness"
        )
        pars.add_argument(
            "--r_subdivs_th", type=float, default=1, help="Minor Line thickness"
        )
        pars.add_argument("--a_divs", type=int, default=24, help="Angle Divisions")
        pars.add_argument(
            "--a_divs_cent", type=int, default=4, help="Angle Divisions at Centre"
        )
        pars.add_argument(
            "--a_subdivs", type=int, default=1, help="Angcular Subdivisions"
        )
        pars.add_argument(
            "--a_subdivs_cent", type=int, default=2, help="Angular Subdivisions end"
        )
        pars.add_argument(
            "--a_divs_th", type=float, default=2, help="Major Angular thickness"
        )
        pars.add_argument(
            "--a_subdivs_th", type=float, default=1, help="Minor Angular thickness"
        )
        pars.add_argument(
            "--c_dot_dia", type=float, default=5.0, help="Diameter of Centre Dot"
        )
        pars.add_argument(
            "--a_labels", default="none", help="The kind of labels to apply"
        )
        pars.add_argument(
            "--a_label_size", type=int, default=18, help="Pixel size of the labels"
        )
        pars.add_argument(
            "--a_label_outset", type=float, default=24, help="Label Radial outset"
        )

    def generate(self):
        self.options.dr = self.svg.unittouu(str(self.options.dr) + "px")
        self.options.r_divs_th = self.svg.unittouu(str(self.options.r_divs_th) + "px")
        self.options.r_subdivs_th = self.svg.unittouu(
            str(self.options.r_subdivs_th) + "px"
        )
        self.options.a_divs_th = self.svg.unittouu(str(self.options.a_divs_th) + "px")
        self.options.a_subdivs_th = self.svg.unittouu(
            str(self.options.a_subdivs_th) + "px"
        )
        self.options.c_dot_dia = self.svg.unittouu(str(self.options.c_dot_dia) + "px")
        self.options.a_label_size = self.svg.unittouu(
            str(self.options.a_label_size) + "px"
        )
        self.options.a_label_outset = self.svg.unittouu(
            str(self.options.a_label_outset) + "px"
        )

        # Embed grid in group
        grid = Group.new("GridPolar:R{0.r_divs}:A{0.a_divs}".format(self.options))

        (pos_x, pos_y) = self.svg.namedview.center
        grid.transform.add_translate(pos_x, pos_y)

        dr = self.options.dr  # Distance between neighbouring circles
        dtheta = (
            2 * pi / self.options.a_divs_cent
        )  # Angular change between adjacent radial lines at centre
        rmax = self.options.r_divs * dr

        # Create SVG circles
        for i in range(1, self.options.r_divs + 1):
            draw_circle(
                i * dr,
                0,
                0,  # major div circles
                self.options.r_divs_th,
                "none",
                "MajorDivCircle" + str(i) + ":R" + str(i * dr),
                grid,
            )

            if self.options.r_log:  # logarithmic subdivisions
                for j in range(2, self.options.r_subdivs):
                    draw_circle(
                        i * dr
                        - (1 - log(j, self.options.r_subdivs))
                        * dr,  # minor div circles
                        0,
                        0,
                        self.options.r_subdivs_th,
                        "none",
                        "MinorDivCircle" + str(i) + ":Log" + str(j),
                        grid,
                    )
            else:  # linear subdivs
                for j in range(1, self.options.r_subdivs):
                    draw_circle(
                        i * dr - j * dr / self.options.r_subdivs,  # minor div circles
                        0,
                        0,
                        self.options.r_subdivs_th,
                        "none",
                        "MinorDivCircle" + str(i) + ":R" + str(i * dr),
                        grid,
                    )

        if (
            self.options.a_divs == self.options.a_divs_cent
        ):  # the lines can go from the centre to the edge
            for i in range(0, self.options.a_divs):
                draw_line(
                    0,
                    0,
                    rmax * sin(i * dtheta),
                    rmax * cos(i * dtheta),
                    self.options.a_divs_th,
                    "RadialGridline" + str(i),
                    grid,
                )

        else:  # we need separate lines
            for i in range(
                0, self.options.a_divs_cent
            ):  # lines that go to the first circle
                draw_line(
                    0,
                    0,
                    dr * sin(i * dtheta),
                    dr * cos(i * dtheta),
                    self.options.a_divs_th,
                    "RadialGridline" + str(i),
                    grid,
                )

            dtheta = (
                2 * pi / self.options.a_divs
            )  # work out the angle change for outer lines

            for i in range(
                0, self.options.a_divs
            ):  # lines that go from there to the edge
                draw_line(
                    dr * sin(i * dtheta + pi / 2.0),
                    dr * cos(i * dtheta + pi / 2.0),
                    rmax * sin(i * dtheta + pi / 2.0),
                    rmax * cos(i * dtheta + pi / 2.0),
                    self.options.a_divs_th,
                    "RadialGridline" + str(i),
                    grid,
                )

        if self.options.a_subdivs > 1:  # draw angular subdivs
            for i in range(0, self.options.a_divs):  # for each major division
                for j in range(1, self.options.a_subdivs):  # draw the subdivisions
                    angle = (
                        i * dtheta - j * dtheta / self.options.a_subdivs + pi / 2.0
                    )  # the angle of the subdivion line
                    draw_line(
                        dr * self.options.a_subdivs_cent * sin(angle),
                        dr * self.options.a_subdivs_cent * cos(angle),
                        rmax * sin(angle),
                        rmax * cos(angle),
                        self.options.a_subdivs_th,
                        "RadialMinorGridline" + str(i),
                        grid,
                    )

        if self.options.c_dot_dia != 0:  # if a non-zero diameter, draw the centre dot
            draw_circle(
                self.options.c_dot_dia / 2.0, 0, 0, 0, "#000000", "CentreDot", grid
            )

        if self.options.a_labels == "deg":
            label_radius = rmax + self.options.a_label_outset  # radius of label centres
            label_size = self.options.a_label_size
            numeral_size = (
                0.73 * label_size
            )  # numerals appear to be 0.73 the height of the nominal pixel size of the font in "Sans"

            for i in range(
                0, self.options.a_divs
            ):  # self.options.a_divs): #radial line labels
                draw_label(
                    sin(i * dtheta + pi / 2.0)
                    * label_radius,  # 0 at the RHS, mathematical style
                    cos(i * dtheta + pi / 2.0) * label_radius
                    + numeral_size / 2.0,  # centre the text vertically
                    str(i * 360 / self.options.a_divs),
                    label_size,
                    "Label" + str(i),
                    grid,
                )

        return grid


if __name__ == "__main__":
    GridPolar().run()



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2007 Terry Brown, terry_n_brown@yahoo.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

from math import atan2, degrees

import inkex
from inkex import ClipPath, Filter


class InsetShadow(inkex.EffectExtension):
    """Generate a 3d edge"""

    def add_arguments(self, pars):
        pars.add_argument(
            "--angle",
            type=float,
            default=45.0,
            help="angle of illumination, clockwise, 45 = upper right",
        )
        pars.add_argument(
            "--stddev", type=float, default=5.0, help="Gaussian Blur stdDeviation"
        )
        pars.add_argument(
            "--blurheight", type=float, default=2.0, help="Gaussian Blur height"
        )
        pars.add_argument(
            "--blurwidth", type=float, default=2.0, help="Gaussian Blur width"
        )
        pars.add_argument("--shades", type=int, default=2, help="Number of shades")
        pars.add_argument("--bw", type=inkex.Boolean, help="Black and white")
        pars.add_argument(
            "--thick", type=float, default=10.0, help="stroke-width for pieces"
        )

    def angle_between(self, start, end, angle):
        """Return true if angle (degrees, clockwise, 0 = up/north) is between
        angles start and end"""

        def f(x):
            """Add 360 to x if x is less than 0"""
            if x < 0:
                return x + 360.0
            return x

        # rotate all inputs by start, => start = 0
        a = f(f(angle) - f(start))
        e = f(f(end) - f(start))
        return a < e

    def effect(self):
        """Check each internode to see if it's in one of the wedges
        for the current shade.  shade is a floating point 0-1 white-black"""
        # size of a wedge for shade i, wedges come in pairs
        delta = 360.0 / self.options.shades / 2.0
        for node in self.svg.selection.filter(inkex.PathElement):
            array = node.path.to_arrays()
            group = None
            filt = None
            for shade in range(0, self.options.shades):
                if self.options.bw and 0 < shade < self.options.shades - 1:
                    continue
                start = [self.options.angle - delta * (shade + 1)]
                end = [self.options.angle - delta * shade]
                start.append(self.options.angle + delta * shade)
                end.append(self.options.angle + delta * (shade + 1))
                level = float(shade) / float(self.options.shades - 1)
                last = []
                result = []
                for cmd, params in array:
                    if cmd == "Z":
                        last = []
                        continue
                    if last:
                        if cmd == "V":
                            point = [last[0], params[-2:][0]]
                        elif cmd == "H":
                            point = [params[-2:][0], last[1]]
                        else:
                            point = params[-2:]
                        ang = degrees(atan2(point[0] - last[0], point[1] - last[1]))
                        if self.angle_between(
                            start[0], end[0], ang
                        ) or self.angle_between(start[1], end[1], ang):
                            result.append(("M", last))
                            result.append((cmd, params))
                        ref = point
                    else:
                        ref = params[-2:]
                    last = ref
                if result:
                    if group is None:
                        group, filt = self.get_group(node)
                    new_node = group.add(node.copy())
                    new_node.path = result
                    new_node.style = "fill:none;stroke-opacity:1;stroke-width:10"
                    new_node.style += filt
                    col = 255 - int(255.0 * level)
                    new_node.style["stroke"] = inkex.Color((col, col, col))

    def get_group(self, node):
        """
        make a clipped group, clip with clone of original, clipped group
        include original and group of paths.
        """
        defs = self.svg.defs
        clip = defs.add(ClipPath())
        new_node = clip.add(node.copy())
        clip_group = node.getparent().add(inkex.Group())
        group = clip_group.add(inkex.Group())
        clip_group.set("clip-path", clip.get_id(as_url=2))

        # make a blur filter reference by the style of each path
        filt = defs.add(
            Filter(
                x="-0.5",
                y="-0.5",
                height=str(self.options.blurheight),
                width=str(self.options.blurwidth),
            )
        )

        filt.add_primitive("feGaussianBlur", stdDeviation=self.options.stddev)
        return group, inkex.Style(filter=filt.get_id(as_url=2))


if __name__ == "__main__":
    InsetShadow().run()



#!/usr/bin/env python3
"""
Convert PNG to WebP using Raster Output extension.
"""

import inkex


class WebpOutput(inkex.RasterOutputExtension):
    def add_arguments(self, pars):
        pars.add_argument("--tab")
        pars.add_argument("--quality", type=int, default=80)
        pars.add_argument("--speed", type=int, default=0)
        pars.add_argument("--lossless", type=inkex.Boolean, default=True)

    def save(self, stream):
        self.img.save(
            stream,
            format="webp",
            quality=self.options.quality,
            lossless=self.options.lossless,
            method=self.options.speed,
        )


if __name__ == "__main__":
    WebpOutput().run()



# coding=utf-8
#
# Copyright (C) 2018 Martin Owens <doctormo@gmail.com>
# Copyright (C) 2023 Jonathan Neuhauser <jonathan.neuhauser@outlook.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
"""Paths module.

Most of the functions derivative, unit_tangent, curvature, point, split, length, ilength
for the individual path commands are ported from
https://github.com/mathandy/svgpathtools/ (MIT licensed)
"""

from typing import Union

from ..transforms import ComplexLike
from .interfaces import (
    PathCommand,
    AbsolutePathCommand,
    RelativePathCommand,
    LengthSettings,
    ILengthSettings,
)
from .lines import Line, line, Move, move, ZoneClose, zoneClose, Horz, horz, Vert, vert
from .curves import curve, Curve, smooth, Smooth
from .quadratic import quadratic, Quadratic, tepidQuadratic, TepidQuadratic
from .arc import Arc, arc, arc_to_path, matprod, rotmat, applymat, norm
from .path import CubicSuperPath, Path, InvalidPath

import numpy as np

np.seterr(invalid="raise")


# definitions that can't be inside the class due to circular dependencies
def to_curve(self, prev: ComplexLike, prev_prev: ComplexLike = 0) -> Curve:
    """Convert command to :py:class:`Curve`

    Curve().to_curve() returns a copy
    """
    return Curve(*self.ccurve_points(0 + 0j, complex(prev), complex(prev_prev)))


def to_line(self, prev: ComplexLike) -> Line:
    """Converts this segment to a line (copies if already a line)"""
    return Line(self.cend_point(0, complex(prev)))


PathCommand.to_curve = to_curve  # type: ignore
PathCommand.to_line = to_line  # type: ignore

PathCommand._letter_to_class = {  # pylint: disable=protected-access
    "M": Move,
    "L": Line,
    "V": Vert,
    "H": Horz,
    "A": Arc,
    "C": Curve,
    "S": Smooth,
    "Z": ZoneClose,
    "Q": Quadratic,
    "T": TepidQuadratic,
    "m": move,
    "l": line,
    "v": vert,
    "h": horz,
    "a": arc,
    "c": curve,
    "s": smooth,
    "z": zoneClose,
    "q": quadratic,
    "t": tepidQuadratic,
}


# All the names that get added to the inkex API itself.
__all__ = (
    "Path",
    "CubicSuperPath",
    "PathCommand",
    "AbsolutePathCommand",
    "RelativePathCommand",
    # Path commands:
    "Line",
    "line",
    "Move",
    "move",
    "ZoneClose",
    "zoneClose",
    "Horz",
    "horz",
    "Vert",
    "vert",
    "Curve",
    "curve",
    "Smooth",
    "smooth",
    "Quadratic",
    "quadratic",
    "TepidQuadratic",
    "tepidQuadratic",
    "Arc",
    "arc",
    # errors
    "InvalidPath",
    # structs
    "LengthSettings",
    "ILengthSettings",
)



# coding=utf-8
#
# Copyright (C) 2018-2019 Martin Owens
#               2019 Thomas Holder
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
#
"""
Testing module. See :ref:`unittests` for details.
"""

import os
import re
import sys
import shutil
import tempfile
import hashlib
import random
import uuid
import io
from typing import List, Union, Tuple, Type, TYPE_CHECKING

from io import BytesIO, StringIO
import xml.etree.ElementTree as xml

from unittest import TestCase as BaseCase
from inkex.base import InkscapeExtension
from inkex.extensions import OutputExtension

from .. import Transform, load_svg, SvgDocumentElement
from ..utils import to_bytes
from .xmldiff import xmldiff
from .mock import MockCommandMixin, Capture

if TYPE_CHECKING:
    from .filters import Compare

COMPARE_DELETE, COMPARE_CHECK, COMPARE_WRITE, COMPARE_OVERWRITE = range(4)


class NoExtension(InkscapeExtension):  # pylint: disable=too-few-public-methods
    """Test case must specify 'self.effect_class' to assertEffect."""

    def __init__(self, *args, **kwargs):  # pylint: disable=super-init-not-called
        raise NotImplementedError(self.__doc__)

    def run(self, args=None, output=None):
        """Fake run"""


class TestCase(MockCommandMixin, BaseCase):
    """
    Base class for all effects tests, provides access to data_files and
    test_without_parameters
    """

    effect_class = NoExtension  # type: Type[InkscapeExtension]
    effect_name = property(lambda self: self.effect_class.__module__)

    # If set to true, the output is not expected to be the stdout SVG document, but
    # rather text or a message sent to the stderr, this is highly weird. But sometimes
    # happens.
    stderr_output = False
    stdout_protect = True
    stderr_protect = True

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self._temp_dir = None
        self._effect = None

    def setUp(self):  # pylint: disable=invalid-name
        """Make sure every test is seeded the same way"""
        self._effect = None
        super().setUp()
        random.seed(0x35F)

    def tearDown(self):
        super().tearDown()
        if self._temp_dir and os.path.isdir(self._temp_dir):
            shutil.rmtree(self._temp_dir)

    @classmethod
    def __file__(cls):
        """Create a __file__ property which acts much like the module version"""
        return os.path.abspath(sys.modules[cls.__module__].__file__)

    @classmethod
    def _testdir(cls):
        """Get's the folder where the test exists (so data can be found)"""
        return os.path.dirname(cls.__file__())

    @classmethod
    def rootdir(cls):
        """Return the full path to the extensions directory"""
        return os.path.dirname(cls._testdir())

    @classmethod
    def datadir(cls):
        """Get the data directory (can be over-ridden if needed)"""
        return os.path.join(cls._testdir(), "data")

    @property
    def tempdir(self):
        """Generate a temporary location to store files"""
        if self._temp_dir is None:
            self._temp_dir = os.path.realpath(tempfile.mkdtemp(prefix="inkex-tests-"))
        if not os.path.isdir(self._temp_dir):
            raise IOError("The temporary directory has disappeared!")
        return self._temp_dir

    def temp_file(
        self, prefix="file-", template="{prefix}{name}{suffix}", suffix=".tmp"
    ):
        """Generate the filename of a temporary file"""
        filename = template.format(prefix=prefix, suffix=suffix, name=uuid.uuid4().hex)
        return os.path.join(self.tempdir, filename)

    @classmethod
    def data_file(cls, filename, *parts, check_exists=True):
        """Provide a data file from a filename, can accept directories as arguments.

        .. versionchanged:: 1.2
            ``check_exists`` parameter added
        """
        if os.path.isabs(filename):
            # Absolute root was passed in, so we trust that (it might be a tempdir)
            full_path = os.path.join(filename, *parts)
        else:
            # Otherwise we assume it's relative to the test data dir.
            full_path = os.path.join(cls.datadir(), filename, *parts)

        if not os.path.isfile(full_path) and check_exists:
            raise IOError(f"Can't find test data file: {full_path}")
        return full_path

    @property
    def empty_svg(self):
        """Returns a common minimal svg file"""
        return self.data_file("svg", "default-inkscape-SVG.svg")

    def assertAlmostTuple(self, found, expected, precision=8, msg=""):  # pylint: disable=invalid-name
        """
        Floating point results may vary with computer architecture; use
        assertAlmostEqual to allow a tolerance in the result.
        """
        self.assertEqual(len(found), len(expected), msg)
        for fon, exp in zip(found, expected):
            self.assertAlmostEqual(fon, exp, precision, msg)

    def assertEffectEmpty(self, effect, **kwargs):  # pylint: disable=invalid-name
        """Assert calling effect without any arguments"""
        self.assertEffect(effect=effect, **kwargs)

    def assertEffect(self, *filename, **kwargs):  # pylint: disable=invalid-name
        """Assert an effect, capturing the output to stdout.

        :py:attr:`filename` should point to a starting svg document, default is empty_svg
        """
        if filename:
            data_file = self.data_file(*filename)
        else:
            data_file = self.empty_svg

        os.environ["DOCUMENT_PATH"] = data_file
        args = [data_file] + list(kwargs.pop("args", []))
        args += [f"--{kw[0]}={kw[1]}" for kw in kwargs.items()]

        effect = kwargs.pop("effect", self.effect_class)()

        # Output is redirected to this string io buffer
        if self.stderr_output:
            with Capture("stderr") as stderr:
                effect.run(args, output=BytesIO())
                effect.test_output = stderr
        else:
            output = BytesIO()
            with Capture(
                "stdout", kwargs.get("stdout_protect", self.stdout_protect)
            ) as stdout:
                with Capture(
                    "stderr", kwargs.get("stderr_protect", self.stderr_protect)
                ) as stderr:
                    effect.run(args, output=output)
                    self.assertEqual(
                        "", stdout.getvalue(), "Extra print statements detected"
                    )
                    self.assertEqual(
                        "", stderr.getvalue(), "Extra error or warnings detected"
                    )
            effect.test_output = output

        if os.environ.get("FAIL_ON_DEPRECATION", False):
            warnings = getattr(effect, "warned_about", set())
            effect.warned_about = set()  # reset for next test
            self.assertFalse(warnings, "Deprecated API is still being used!")

        return effect

    # pylint: disable=invalid-name
    def assertDeepAlmostEqual(self, first, second, places=None, msg=None, delta=None):
        """Asserts that two objects, possible nested lists, are almost equal."""
        if delta is None and places is None:
            places = 7
        if isinstance(first, (list, tuple)):
            assert len(first) == len(second)
            for f, s in zip(first, second):
                self.assertDeepAlmostEqual(f, s, places, msg, delta)
        else:
            self.assertAlmostEqual(first, second, places, msg, delta)

    def assertTransformEqual(self, lhs, rhs, places=7):
        """Assert that two transform expressions evaluate to the same
        transformation matrix.

        .. versionadded:: 1.1
        """
        self.assertAlmostTuple(
            tuple(Transform(lhs).to_hexad()), tuple(Transform(rhs).to_hexad()), places
        )

    # pylint: enable=invalid-name

    @property
    def effect(self):
        """Generate an effect object"""
        if self._effect is None:
            self._effect = self.effect_class()
        return self._effect

    def import_string(self, string, *args) -> SvgDocumentElement:
        """Runs a string through an import extension, with optional arguments
        provided as "--arg=value" arguments"""
        stream = io.BytesIO(string.encode())
        reader = self.effect_class()
        out = io.BytesIO()
        reader.parse_arguments([*args])
        reader.options.input_file = stream
        reader.options.output = out
        reader.load_raw()
        reader.save_raw(reader.effect())
        out.seek(0)
        decoded = out.read().decode("utf-8")
        document = load_svg(decoded)
        return document

    def export_svg(self, document, *args) -> str:
        """Runs a svg through an export extension, with optional arguments
        provided as "--arg=value" arguments"""
        assert isinstance(self, OutputExtension)
        output = StringIO()
        writer = self.effect_class()
        writer.parse_arguments([*args])
        writer.svg = document.getroot()
        writer.document = document
        writer.effect()
        writer.save(output)
        output.seek(0)
        return output.read()


class InkscapeExtensionTestMixin:
    """Automatically setup self.effect for each test and test with an empty svg"""

    def setUp(self):  # pylint: disable=invalid-name
        """Check if there is an effect_class set and create self.effect if it is"""
        super().setUp()
        if self.effect_class is None:
            self.skipTest("self.effect_class is not defined for this this test")

    def test_default_settings(self):
        """Extension works with empty svg file"""
        self.effect.run([self.empty_svg])


class ComparisonMeta(type):
    """Metaclass for ComparisonMixin which creates parametrized tests that can be run
    independently. See :class:`~inkex.tester.ComparisonMixin` for details.

    ..versionadded :: 1.4
    """

    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)
        if name == "ComparisonMixin":
            return  # don't execute on the base class
        _compare_file = cls.compare_file
        _comparisons = cls.comparisons
        if hasattr(cls, "comparisons_cmpfile_dict"):
            _comparisons = cls.comparisons_cmpfile_dict.keys()

            def get_compare_cmpfile(self, args, addout=None):
                return self.data_file("refs", cls.comparisons_cmpfile_dict[args])

            setattr(cls, "get_compare_cmpfile", get_compare_cmpfile)

        test_method_name = f"test_all_comparisons"
        if hasattr(cls, test_method_name):
            return  # Custom test logic, don't touch

        append = isinstance(_compare_file, (list, tuple))
        compare_file = _compare_file if append else [_compare_file]
        try:
            for file in compare_file:
                for comparison in _comparisons:
                    test_method_name = f"test_comparison_{comparison}_{file}"
                    setattr(
                        cls,
                        test_method_name,
                        ComparisonMeta.create_test_method(
                            comparison,
                            file,
                            os.path.basename(file) if append else None,
                        ),
                    )
        except TypeError:
            # If the compare_file or compare_
            test_method_name = f"test_all_comparisons"

            def test_method(self):
                if not isinstance(self.compare_file, (list, tuple)):
                    self._test_comparisons(self.compare_file)
                else:
                    for compare_file in self.compare_file:
                        self._test_comparisons(
                            compare_file, addout=os.path.basename(compare_file)
                        )

            setattr(cls, test_method_name, test_method)

    @staticmethod
    def create_test_method(comparison, file, addout):
        def _test_method(self):
            self._test_comparison(comparison, file, addout)

        return _test_method


class ComparisonMixin(metaclass=ComparisonMeta):
    """
    This mixin allows to easily specify a set of run-through unit tests for an
    extension, which is specified in :attr:`inkex.tester.TestCase.effect_class`.

    The commandline parameters are passed in :attr:`comparisons`, the input file
    in :attr:`compare_file` (either a list of files, or a single file).

    The :class:`ComparisonMeta` metaclass creates a set of independent unit tests
    out of this data. Behavior notest:

    - The unit tests created are the cross product of :attr:`comparisons` and
      :attr:`compare_file`. If :attr:`compare_file` is a list, the comparison file name
      is suffixed with the current ``compare_file`` name.
    - Optionally, :attr:`comparisons_cmpfile_dict` may be specified as
      ``Dict[Tuple[str], str]`` where the keys are sets of command line parameters and
      the values are the filenames of the output file. This takes precedence over
      :attr:`comparisons`.
    - If any of those values are properties, their values cannot be accessed at test
      collection time and there will only be a single test, ``test_all_comparisons``
      with otherwise identical behavior.
    - If the class overrides ``test_all_comparisons``, no additional tests are
      generated to allow for custom comparison logic.

    To create the comparison files for the unit tests, use the ``EXPORT_COMPARE``
    environment variable.
    """

    compare_file: Union[List[str], Tuple[str], str] = "svg/shapes.svg"
    """This input svg file sent to the extension (if any)"""

    compare_filters = []  # type: List[Compare]
    """The ways in which the output is filtered for comparision (see filters.py)"""

    compare_filter_save = False
    """If true, the filtered output will be saved and only applied to the
    extension output (and not to the reference file)"""

    comparisons = [
        (),
        ("--id=p1", "--id=r3"),
    ]
    """A list of comparison runs, each entry will cause the extension to be run."""

    compare_file_extension = "svg"

    @property
    def _compare_file_extension(self):
        """The default extension to use when outputting check files in COMPARE_CHECK
        mode."""
        if self.stderr_output:
            return "txt"
        return self.compare_file_extension

    def _test_comparisons(self, compare_file, addout=None):
        for args in self.comparisons:
            self._test_comparison(args, compare_file=compare_file, addout=addout)

    def _test_comparison(self, args, compare_file, addout=None):
        self.assertCompare(
            compare_file,
            self.get_compare_cmpfile(args, addout),
            args,
        )

    def assertCompare(self, infile, cmpfile, args, outfile=None):  # pylint: disable=invalid-name
        """
        Compare the output of a previous run against this one.

        Args:
            infile: The filename of the pre-processed svg (or other type of file)
            cmpfile: The filename of the data we expect to get, if not set
                the filename will be generated from the effect name and kwargs.
            args: All the arguments to be passed to the effect run
            outfile: Optional, instead of returning a regular output, this extension
                dumps it's output to this filename instead.

        """
        compare_mode = int(os.environ.get("EXPORT_COMPARE", COMPARE_DELETE))

        effect = self.assertEffect(infile, args=args)

        if cmpfile is None:
            cmpfile = self.get_compare_cmpfile(args)

        if not os.path.isfile(cmpfile) and compare_mode == COMPARE_DELETE:
            raise IOError(
                f"Comparison file {cmpfile} not found, set EXPORT_COMPARE=1 to create "
                "it."
            )

        if outfile:
            if not os.path.isabs(outfile):
                outfile = os.path.join(self.tempdir, outfile)
            self.assertTrue(
                os.path.isfile(outfile), f"No output file created! {outfile}"
            )
            with open(outfile, "rb") as fhl:
                data_a = fhl.read()
        else:
            data_a = effect.test_output.getvalue()

        write_output = None
        if compare_mode == COMPARE_CHECK:
            _file = cmpfile[:-4] if cmpfile.endswith(".out") else cmpfile
            write_output = f"{_file}.{self._compare_file_extension}"
        elif (
            compare_mode == COMPARE_WRITE and not os.path.isfile(cmpfile)
        ) or compare_mode == COMPARE_OVERWRITE:
            write_output = cmpfile

        try:
            if write_output and not os.path.isfile(cmpfile):
                raise AssertionError(f"Check the output: {write_output}")
            with open(cmpfile, "rb") as fhl:
                data_b = self._apply_compare_filters(fhl.read(), False)
            self._base_compare(data_a, data_b, compare_mode)
        except AssertionError:
            if write_output:
                if isinstance(data_a, str):
                    data_a = data_a.encode("utf-8")
                self.write_compare_data(infile, write_output, data_a)
                # This only reruns if the original test failed.
                # The idea here is to make sure the new output file is "stable"
                # Because some tests can produce random changes and we don't
                # want test authors to be too reassured by a simple write.
                if write_output == cmpfile:
                    effect = self.assertEffect(infile, args=args)
                    self._base_compare(data_a, cmpfile, COMPARE_CHECK)
            if not write_output == cmpfile:
                raise

    def write_compare_data(self, infile, outfile, data):
        """Write output"""
        with open(outfile, "wb") as fhl:
            fhl.write(self._apply_compare_filters(data, True))
            print(f"Written output: {outfile}")

    def _base_compare(self, data_a, data_b, compare_mode):
        data_a = self._apply_compare_filters(data_a)

        if (
            isinstance(data_a, bytes)
            and isinstance(data_b, bytes)
            and data_a.startswith(b"<")
            and data_b.startswith(b"<")
        ):
            # Late importing
            diff_xml, delta = xmldiff(data_a, data_b)
            if not delta and compare_mode == COMPARE_DELETE:
                print(
                    "The XML is different, you can save the output using the "
                    "EXPORT_COMPARE envionment variable. Set it to 1 to save a file "
                    "you can check, set it to 3 to overwrite this comparison, setting "
                    "the new data as the correct one.\n"
                )
            diff = "SVG Differences\n\n"
            if os.environ.get("XML_DIFF", False):
                diff = "<- " + diff_xml
            else:
                for x, (value_a, value_b) in enumerate(delta):
                    try:
                        # Take advantage of better text diff in testcase's own asserts.
                        self.assertEqual(value_a, value_b)
                    except AssertionError as err:
                        diff += f" {x}. {str(err)}\n"
            self.assertTrue(delta, diff)
        else:
            # compare any content (non svg)
            self.assertEqual(data_a, data_b)

    def _apply_compare_filters(self, data, is_saving=None):
        data = to_bytes(data)
        # Applying filters flips depending if we are saving the filtered content
        # to disk, or filtering during the test run. This is because some filters
        # are destructive others are useful for diagnostics.
        if is_saving is self.compare_filter_save or is_saving is None:
            for cfilter in self.compare_filters:
                data = cfilter(data)
        return data

    def get_compare_cmpfile(self, args, addout=None):
        """Generate an output file for the arguments given"""
        if addout is not None:
            args = list(args) + [str(addout)]
        opstr = (
            "__".join(args)
            .replace(self.tempdir, "TMP_DIR")
            .replace(self.datadir(), "DAT_DIR")
        )
        opstr = re.sub(r"[^\w-]", "__", opstr)
        if opstr:
            if len(opstr) > 127:
                # avoid filename-too-long error
                opstr = hashlib.md5(opstr.encode("latin1")).hexdigest()
            opstr = "__" + opstr
        return self.data_file(
            "refs", f"{self.effect_name}{opstr}.out", check_exists=False
        )



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (C) 2022 Jonathan Neuhauser (jonathan.neuhauser@outlook.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

"""Grammar for parsing HPGL input files"""

import pyparsing as pp
from hpgl_input_sm import HPGLStateMachine

pp.ParserElement.enablePackrat()
pp.enable_all_warnings()

integer = pp.pyparsing_common.integer
sinteger = pp.pyparsing_common.signed_integer
flt = pp.pyparsing_common.number
comma = pp.Optional(pp.Suppress(","))

cpair = flt("X*") + comma + flt("Y*")


def build_vector_parsers(stm: HPGLStateMachine):
    """Parsers of the vector group"""
    # Lines
    line_command = pp.Group(
        (
            (pp.Literal("PA") | "PR" | "PD" | "PU")("key")
            + pp.Opt(
                cpair
                + pp.ZeroOrMore(comma + cpair)
                + pp.Opt(comma + sinteger("leftover*"))
            )
        ).add_parse_action(stm.line_command)
    )
    # Circles and arcs
    ci_command = pp.Group(
        (
            "CI" + sinteger("Radius") + pp.Opt(comma + flt("Chord_angle"))
        ).add_parse_action(stm.ci_command)
    )

    arc_command = pp.Group(
        (
            (pp.Literal("AA") | "AR")("key")
            + cpair
            + comma
            + sinteger("sweep")
            + pp.Opt(comma + flt("Chord_angle*"))
        ).add_parse_action(stm.arc_command)
    )
    at_command = pp.Group(
        (
            "AT" + cpair + comma + cpair + pp.Opt(comma + flt("Chord_angle*"))
        ).add_parse_action(stm.at_command)
    )

    # Beziers
    bezier = cpair + comma + cpair + comma + cpair
    bezier_command = pp.Group(
        (
            (pp.Literal("BR") | "BZ")("key")
            + bezier("B*")
            + pp.ZeroOrMore(comma + bezier("B*"))
        ).add_parse_action(stm.bezier_command)
    )
    # Polyline encoded
    pe_command = pp.Group(
        (
            "PE" + pp.Regex("[^;]*").setWhitespaceChars("")("data") + pp.Suppress(";")
        ).add_parse_action(stm.polyline_encoded)
    )

    return (
        line_command
        | ci_command
        | arc_command
        | bezier_command
        | at_command
        | pe_command
    )


def build_polygon_parsers(stm: HPGLStateMachine):
    """Build parser object from the polygon group"""
    pm_command = pp.Group(
        ("PM" + pp.Opt(pp.Word("012", exact=1), default="0")("value")).add_parse_action(
            stm.pm_command
        )
    )
    rectangle_command = pp.Group(
        ((pp.Literal("EA") | "ER" | "RA" | "RR")("key") + cpair).add_parse_action(
            stm.rectangle_command
        )
    )
    polygon_command = pp.Group(
        (
            (
                pp.Literal("FP")("key")
                + pp.Opt(pp.Word("01", exact=1), default="0")("fillmode")
            )
            | pp.Literal("EP")("key")
        ).add_parse_action(stm.edge_fill_polygon)
    )
    wedge_command = pp.Group(
        (
            (pp.Literal("EW") | "WG")("key")
            + sinteger("radius")
            + comma
            + flt("start_angle")
            + comma
            + flt("sweep")
            + pp.Opt(comma + flt("Chord_angle*"))
        ).add_parse_action(stm.wedge_command)
    )
    return pm_command | rectangle_command | polygon_command | wedge_command


def build_linefill_parsers(stm: HPGLStateMachine):
    """Build parsers from the Line and Fill Attributes Group"""

    kind = pp.Word("123", exact=1)("kind*")
    la_command = pp.Group(
        "LA"
        + pp.Opt(kind + comma + integer("value*"))
        + (comma + kind + comma + integer("value*")) * (0, 2)
    )
    lt_command = pp.Group(
        "LT"
        + pp.Opt(
            integer("linetype")
            + pp.Opt(comma + flt("pattern_length") + pp.Opt(comma + flt("mode")))
        )
    )
    pw_command = pp.Group("PW" + pp.Opt(flt("width") + pp.Opt(comma + integer("pen"))))
    # maybe we can combine those two?
    ft_command = pp.Group(
        "FT"
        + pp.Opt(
            (pp.Literal("1") | "2" | "3" | "4" | "10" | "11" | "21" | "22")("type")
            + pp.Opt(comma + flt("option1") + pp.Opt(comma + flt("option2")))
        )
    )
    tr_command = pp.Group("TR" + pp.Opt(pp.Literal("0") | "1", "1")("transparency"))
    ul_command = pp.Group(
        "UL" + pp.Opt(integer("index") + (comma + flt("gap*")) * (0, 20))
    )
    wu_command = pp.Group("WU" + pp.Opt(pp.Word("01", exact=1)("units")))
    sp_command = pp.Group("SP" + pp.Opt(integer("pen"), 0))
    return (
        sp_command
        | la_command
        | lt_command
        | pw_command
        | ft_command
        | tr_command
        | ul_command
        | wu_command
    ).add_parse_action(stm.style_command)


def build_configuration_parser(stm: HPGLStateMachine):
    """Commands of the configurtion group"""
    co_command = pp.Literal("CO") + pp.Suppress('"') + ... + pp.Suppress('"')
    in_command = pp.Literal("IN").add_parse_action(stm.initialize)
    ip_command = pp.Group(
        (pp.Literal("IP") | "IR")("key") + pp.Opt(cpair + pp.Opt(comma + cpair))
    )
    ro_command = pp.Group(
        "RO" + pp.Optional(pp.Literal("0") | "90" | "180" | "270", "0")("angle")
    )
    iw_command = pp.Group("IW" + pp.Opt(cpair + comma + cpair))
    sc_command = pp.Group(
        pp.Literal("SC")
        + (
            pp.Optional(
                flt("Xmin")
                + comma
                + flt("Xmax")
                + comma
                + flt("Ymin")
                + comma
                + flt("Ymax")
                + pp.Optional(
                    (comma + (pp.Literal("0") | "2")("type"))
                    | (
                        comma
                        + pp.Literal("1")("type")
                        + pp.Optional(comma + flt("left") + comma + flt("bottom"))
                    )
                )
            )
        )
    )

    return (in_command | co_command) | (
        ip_command | sc_command | ro_command | iw_command
    ).add_parse_action(stm.transform_command)


def build_parser(stm: HPGLStateMachine):
    """Assemble the command groups"""

    command = (
        build_configuration_parser(stm)
        | build_vector_parsers(stm)
        | build_polygon_parsers(stm)
        | build_linefill_parsers(stm)
        # Commands we don't understand
        | pp.Group(pp.Word(pp.alphas, exact=2) + pp.Opt(pp.Word(pp.nums + ",. ")))("u*")
    )
    # The parse actions of the documents are final cleanup
    document = pp.ZeroOrMore(command + pp.Opt(pp.Suppress(";"))).add_parse_action(
        stm.finalize_path
    )
    pp.autoname_elements()
    return document



#!/usr/bin/env python3
"""
Testing tool for running queries.
"""

import sys
import json

from import_sources import RemoteSource


def run():
    RemoteSource.load("sources")
    source = RemoteSource.sources[sys.argv[1]]("/tmp")

    for item in source.search(sys.argv[2]):
        if callable(item):
            print(f"Next Page {item}")
        else:
            print(json.dumps(item, indent=2))


if __name__ == "__main__":
    run()



"""
VNImageElement
"""

import base64
from dataclasses import dataclass
from io import BytesIO
from typing import List, Optional, Tuple, Union

import inkex
from PIL import Image

from .base import VNBaseElement


@dataclass
class VNImageElement(VNBaseElement):
    """
    Holds imageData as base64 texts.

    transform contains matrix (old format).
    """

    imageData: str
    transform: Optional[List[float]]
    cropRect: Optional[Tuple[Tuple[float, float], Tuple[float, float]]]

    def image_format(self) -> str:
        """Detect the image format of b64 encoded image."""
        binary_data = base64.b64decode(self.imageData)
        image = Image.open(BytesIO(binary_data))
        return image.format

    def image_dimension(self) -> Tuple[int, int]:
        """Detect the dimension format of b64 encoded image."""
        binary_data = base64.b64decode(self.imageData)
        image = Image.open(BytesIO(binary_data))
        return image.width, image.height

    def convert_crop_rect(self) -> Union[inkex.Rectangle, None]:
        if self.cropRect is not None:
            width, height = self.cropRect[1]
            x, y = self.cropRect[0]

            clip_rect = inkex.Rectangle.new(x, y, width, height)
            clip_rect.label = f"{self.name}_crop"
            return clip_rect



# coding=utf-8
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# pylint: disable=invalid-name,unused-argument
"""Deprecated bezmisc API"""

from inkex.deprecated import deprecate
from inkex import bezier

bezierparameterize = deprecate(bezier.bezierparameterize)
linebezierintersect = deprecate(bezier.linebezierintersect)
bezierpointatt = deprecate(bezier.bezierpointatt)
bezierslopeatt = deprecate(bezier.bezierslopeatt)
beziertatslope = deprecate(bezier.beziertatslope)
tpoint = deprecate(bezier.tpoint)
beziersplitatt = deprecate(bezier.beziersplitatt)
pointdistance = deprecate(bezier.pointdistance)
Gravesen_addifclose = deprecate(bezier.addifclose)
balf = deprecate(bezier.balf)
bezierlengthSimpson = deprecate(bezier.bezierlength)
beziertatlength = deprecate(bezier.beziertatlength)
bezierlength = bezierlengthSimpson


@deprecate
def Simpson(func, a, b, n_limit, tolerance):
    """bezier.simpson(a, b, n_limit, tolerance, balf_arguments)"""
    raise AttributeError(
        """Because bezmisc.Simpson used global variables, it's not possible to
        call the replacement code automatically. In fact it's unlikely you were
        using the code or functionality you think you were since it's a highly
        broken way of writing python."""
    )



# coding=utf-8
#
# Copyright (C) 2010 Nick Drobchenko, nick@cnc-club.ru
# Copyright (C) 2005 Aaron Spike, aaron@ekips.org
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# pylint: disable=invalid-name,too-many-locals
#
"""
Bezier calculations
"""

import cmath
import math

import numpy

from .transforms import DirectedLineSegment
from .localization import inkex_gettext as _

# bez = ((bx0,by0),(bx1,by1),(bx2,by2),(bx3,by3))


def pointdistance(point_a, point_b):
    """The straight line distance between two points"""
    return math.sqrt(
        ((point_b[0] - point_a[0]) ** 2) + ((point_b[1] - point_a[1]) ** 2)
    )


def between_point(point_a, point_b, time=0.5):
    """Returns the point between point a and point b"""
    return point_a[0] + time * (point_b[0] - point_a[0]), point_a[1] + time * (
        point_b[1] - point_a[1]
    )


def percent_point(point_a, point_b, percent=50.0):
    """Returns between_point but takes percent instead of 0.0-1.0"""
    return between_point(point_a, point_b, percent / 100.0)


def root_wrapper(root_a, root_b, root_c, root_d):
    """Get the Cubic function, moic formular of roots, simple root"""
    if root_a:
        # Monics formula, see
        # http://en.wikipedia.org/wiki/Cubic_function#Monic_formula_of_roots
        mono_a, mono_b, mono_c = (root_b / root_a, root_c / root_a, root_d / root_a)
        m = 2.0 * mono_a**3 - 9.0 * mono_a * mono_b + 27.0 * mono_c
        k = mono_a**2 - 3.0 * mono_b
        n = m**2 - 4.0 * k**3
        w1 = -0.5 + 0.5 * cmath.sqrt(-3.0)
        w2 = -0.5 - 0.5 * cmath.sqrt(-3.0)
        if n < 0:
            m1 = pow(complex((m + cmath.sqrt(n)) / 2), 1.0 / 3)
            n1 = pow(complex((m - cmath.sqrt(n)) / 2), 1.0 / 3)
        else:
            if m + math.sqrt(n) < 0:
                m1 = -pow(-(m + math.sqrt(n)) / 2, 1.0 / 3)
            else:
                m1 = pow((m + math.sqrt(n)) / 2, 1.0 / 3)
            if m - math.sqrt(n) < 0:
                n1 = -pow(-(m - math.sqrt(n)) / 2, 1.0 / 3)
            else:
                n1 = pow((m - math.sqrt(n)) / 2, 1.0 / 3)
        return (
            -1.0 / 3 * (mono_a + m1 + n1),
            -1.0 / 3 * (mono_a + w1 * m1 + w2 * n1),
            -1.0 / 3 * (mono_a + w2 * m1 + w1 * n1),
        )
    if root_b:
        det = root_c**2.0 - 4.0 * root_b * root_d
        if det:
            return (
                (-root_c + cmath.sqrt(det)) / (2.0 * root_b),
                (-root_c - cmath.sqrt(det)) / (2.0 * root_b),
            )
        return (-root_c / (2.0 * root_b),)
    if root_c:
        return (1.0 * (-root_d / root_c),)
    return ()


def bezlenapprx(sp1, sp2):
    """Return the aproximate length between two beziers"""
    return (
        pointdistance(sp1[1], sp1[2])
        + pointdistance(sp1[2], sp2[0])
        + pointdistance(sp2[0], sp2[1])
    )


def cspbezsplit(sp1, sp2, time=0.5):
    """Split a cubic bezier at the time period"""
    m1 = tpoint(sp1[1], sp1[2], time)
    m2 = tpoint(sp1[2], sp2[0], time)
    m3 = tpoint(sp2[0], sp2[1], time)
    m4 = tpoint(m1, m2, time)
    m5 = tpoint(m2, m3, time)
    m = tpoint(m4, m5, time)
    return [[sp1[0][:], sp1[1][:], m1], [m4, m, m5], [m3, sp2[1][:], sp2[2][:]]]


def cspbezsplitatlength(sp1, sp2, length=0.5, tolerance=0.001):
    """Split a cubic bezier at length"""
    bez = (sp1[1][:], sp1[2][:], sp2[0][:], sp2[1][:])
    time = beziertatlength(bez, length, tolerance)
    return cspbezsplit(sp1, sp2, time)


def cspseglength(sp1, sp2, tolerance=0.001):
    """Get cubic bezier segment length"""
    bez = (sp1[1][:], sp1[2][:], sp2[0][:], sp2[1][:])
    return bezierlength(bez, tolerance)


def csplength(csp):
    """Get cubic bezier length"""
    total = 0
    lengths = []
    for sp in csp:
        lengths.append([])
        for i in range(1, len(sp)):
            l = cspseglength(sp[i - 1], sp[i])
            lengths[-1].append(l)
            total += l
    return lengths, total


def bezierparameterize(bez):
    """Return the bezier parameter size
    Converts the bezier parametrisation from the default form
    P(t) = (1-t)³ P_1 + 3(1-t)²t P_2 + 3(1-t)t² P_3 + t³ x_4
    to the a form which can be differentiated more easily
    P(t) = a t³ + b t² + c t + P0

    Args:
        bez (List[Tuple[float, float]]): the Bezier curve. The elements of the list the
            coordinates of the points (in this order): Start point, Start control point,
            End control point, End point.

    Returns:
        Tuple[float, float, float, float, float, float, float, float]:
            the values ax, ay, bx, by, cx, cy, x0, y0
    """
    ((bx0, by0), (bx1, by1), (bx2, by2), (bx3, by3)) = bez
    # parametric bezier
    x0 = bx0
    y0 = by0
    cx = 3 * (bx1 - x0)
    bx = 3 * (bx2 - bx1) - cx
    ax = bx3 - x0 - cx - bx
    cy = 3 * (by1 - y0)
    by = 3 * (by2 - by1) - cy
    ay = by3 - y0 - cy - by

    return ax, ay, bx, by, cx, cy, x0, y0


def linebezierintersect(arg_a, bez):
    """Where a line and bezier intersect"""
    ((lx1, ly1), (lx2, ly2)) = arg_a
    # parametric line
    dd = lx1
    cc = lx2 - lx1
    bb = ly1
    aa = ly2 - ly1

    if aa:
        coef1 = cc / aa
        coef2 = 1
    else:
        coef1 = 1
        coef2 = aa / cc

    ax, ay, bx, by, cx, cy, x0, y0 = bezierparameterize(bez)
    # cubic intersection coefficients
    a = coef1 * ay - coef2 * ax
    b = coef1 * by - coef2 * bx
    c = coef1 * cy - coef2 * cx
    d = coef1 * (y0 - bb) - coef2 * (x0 - dd)

    roots = root_wrapper(a, b, c, d)
    retval = []
    for i in roots:
        if isinstance(i, complex) and i.imag == 0:
            i = i.real
        if not isinstance(i, complex) and 0 <= i <= 1:
            retval.append(bezierpointatt(bez, i))
    return retval


def bezierpointatt(bez, t):
    """Get coords at the given time point along a bezier curve"""
    ax, ay, bx, by, cx, cy, x0, y0 = bezierparameterize(bez)
    x = ax * (t**3) + bx * (t**2) + cx * t + x0
    y = ay * (t**3) + by * (t**2) + cy * t + y0
    return x, y


def bezierslopeatt(bez, t):
    """Get slope at the given time point along a bezier curve
        The slope is computed as (dx, dy) where dx = df_x(t)/dt and dy = df_y(t)/dt.
        Note that for lines P1=P2 and P3=P4, so the slope at the end points is dx=dy=0
        (slope not defined).

    Args:
        bez (List[Tuple[float, float]]): the Bezier curve. The elements of the list the
            coordinates of the points (in this order): Start point, Start control point,
            End control point, End point.
        t (float): time in the interval [0, 1]

    Returns:
        Tuple[float, float]: x and y increment
    """
    ax, ay, bx, by, cx, cy, _, _ = bezierparameterize(bez)
    dx = 3 * ax * (t**2) + 2 * bx * t + cx
    dy = 3 * ay * (t**2) + 2 * by * t + cy
    return dx, dy


def beziertatslope(bez, d):
    """Reverse; get time from slope along a bezier curve"""
    ax, ay, bx, by, cx, cy, _, _ = bezierparameterize(bez)
    (dy, dx) = d
    # quadratic coefficients of slope formula
    if dx:
        slope = 1.0 * (dy / dx)
        a = 3 * ay - 3 * ax * slope
        b = 2 * by - 2 * bx * slope
        c = cy - cx * slope
    elif dy:
        slope = 1.0 * (dx / dy)
        a = 3 * ax - 3 * ay * slope
        b = 2 * bx - 2 * by * slope
        c = cx - cy * slope
    else:
        return []

    roots = root_wrapper(0, a, b, c)
    retval = []
    for i in roots:
        if isinstance(i, complex) and i.imag == 0:
            i = i.real
        if not isinstance(i, complex) and 0 <= i <= 1:
            retval.append(i)
    return retval


def tpoint(p1, p2, t):
    """Linearly interpolate between p1 and p2.

    t = 0.0 returns p1, t = 1.0 returns p2.

    :return: Interpolated point
    :rtype: tuple

    :param p1: First point as sequence of two floats
    :param p2: Second point as sequence of two floats
    :param t: Number between 0.0 and 1.0
    :type t: float
    """
    x1, y1 = p1
    x2, y2 = p2
    return x1 + t * (x2 - x1), y1 + t * (y2 - y1)


def beziersplitatt(bez, t):
    """Split bezier at given time"""
    ((bx0, by0), (bx1, by1), (bx2, by2), (bx3, by3)) = bez
    m1 = tpoint((bx0, by0), (bx1, by1), t)
    m2 = tpoint((bx1, by1), (bx2, by2), t)
    m3 = tpoint((bx2, by2), (bx3, by3), t)
    m4 = tpoint(m1, m2, t)
    m5 = tpoint(m2, m3, t)
    m = tpoint(m4, m5, t)

    return ((bx0, by0), m1, m4, m), (m, m5, m3, (bx3, by3))


def addifclose(bez, l, error=0.001):
    """Gravesen, Add if the line is closed, in-place addition to array l"""
    box = 0
    for i in range(1, 4):
        box += pointdistance(bez[i - 1], bez[i])
    chord = pointdistance(bez[0], bez[3])
    if (box - chord) > error:
        first, second = beziersplitatt(bez, 0.5)
        addifclose(first, l, error)
        addifclose(second, l, error)
    else:
        l[0] += (box / 2.0) + (chord / 2.0)


# balfax, balfbx, balfcx, balfay, balfby, balfcy = 0, 0, 0, 0, 0, 0


def balf(t, args):
    """Bezier Arc Length Function"""
    ax, bx, cx, ay, by, cy = args
    retval = (ax * (t**2) + bx * t + cx) ** 2 + (ay * (t**2) + by * t + cy) ** 2
    return math.sqrt(retval)


def simpson(start, end, maxiter, tolerance, bezier_args):
    """Calculate the length of a bezier curve using Simpson's algorithm:
    http://steve.hollasch.net/cgindex/curves/cbezarclen.html

    Args:
        start (int): Start time (between 0 and 1)
        end (int): End time (between start time and 1)
        maxiter (int): Maximum number of iterations. If not a power of 2, the algorithm
        will behave like the value is set to the next power of 2.
        tolerance (float):  maximum error ratio
        bezier_args (list): arguments as computed by bezierparametrize()

    Returns:
        float: the appoximate length of the bezier curve
    """

    n = 2
    multiplier = (end - start) / 6.0
    endsum = balf(start, bezier_args) + balf(end, bezier_args)
    interval = (end - start) / 2.0
    asum = 0.0
    bsum = balf(start + interval, bezier_args)
    est1 = multiplier * (endsum + (2.0 * asum) + (4.0 * bsum))
    est0 = 2.0 * est1
    # print(multiplier, endsum, interval, asum, bsum, est1, est0)
    while n < maxiter and abs(est1 - est0) > tolerance:
        n *= 2
        multiplier /= 2.0
        interval /= 2.0
        asum += bsum
        bsum = 0.0
        est0 = est1
        for i in range(1, n, 2):
            bsum += balf(start + (i * interval), bezier_args)
            est1 = multiplier * (endsum + (2.0 * asum) + (4.0 * bsum))
    # print(multiplier, endsum, interval, asum, bsum, est1, est0)
    return est1


def bezierlength(bez, tolerance=0.001, time=1.0):
    """Get length of bezier curve"""
    ax, ay, bx, by, cx, cy, _, _ = bezierparameterize(bez)
    return simpson(0.0, time, 4096, tolerance, [3 * ax, 2 * bx, cx, 3 * ay, 2 * by, cy])


def beziertatlength(bez, l=0.5, tolerance=0.001):
    """Get bezier curve time at the length specified"""
    curlen = bezierlength(bez, tolerance, 1.0)
    time = 1.0
    tdiv = time
    targetlen = l * curlen
    diff = curlen - targetlen
    while abs(diff) > tolerance:
        tdiv /= 2.0
        if diff < 0:
            time += tdiv
        else:
            time -= tdiv
        curlen = bezierlength(bez, tolerance, time)
        diff = curlen - targetlen
    return time


def maxdist(bez):
    """Get maximum distance within bezier curve"""
    seg = DirectedLineSegment(bez[0], bez[3])
    return max(seg.distance_to_point(*bez[1]), seg.distance_to_point(*bez[2]))


def cspsubdiv(csp, flat):
    """Sub-divide cubic sub-paths"""
    for sp in csp:
        subdiv(sp, flat)


def subdiv(sp, flat, i=1):
    """sub divide bezier curve"""
    while i < len(sp):
        p0 = sp[i - 1][1]
        p1 = sp[i - 1][2]
        p2 = sp[i][0]
        p3 = sp[i][1]

        bez = (p0, p1, p2, p3)
        mdist = maxdist(bez)
        if mdist <= flat:
            i += 1
        else:
            one, two = beziersplitatt(bez, 0.5)
            sp[i - 1][2] = one[1]
            sp[i][0] = two[2]
            p = [one[2], one[3], two[1]]
            sp[i:1] = [p]


def csparea(csp):
    r"""Get total area of cubic superpath.

    .. hint::

        The results may be slightly inaccurate for paths containing arcs due
        to the loss of accuracy during arc -> cubic bezier conversion.


    The function works as follows: For each subpath,

    #. compute the area of the polygon created by the path's vertices:

       For a line with coordinates :math:`(x_0, y_0)` and :math:`(x_1, y_1)`, the area
       of the trapezoid of its projection on the x axis is given by

       .. math::

           \frac{1}{2} (y_1 + y_0) (x_1 - x_0)

       Summing the contribution of all lines of the polygon yields the polygon's area
       (lines from left to right have a positive contribution, while those right-to
       left have a negative area contribution, canceling out the computed area not
       inside the polygon), so we find (setting :math:`x_{0} = x_N` etc.):

       .. math::

           A = \frac{1}{2} * \sum_{i=1}^N (x_i y_i - x_{i-1} y_{i-1} + x_i y_{i-1}
           - x_{i-1} y_{i})

       The first two terms cancel out in the summation over all points, and the second
       two terms can be regrouped as

       .. math::

           A = \frac{1}{2} * \sum_{i=1}^N x_i (y_{i+1} -y_{i-1})

    #. The contribution by the bezier curve is considered: We compute
       the integral :math:`\int_{x(t=0)}^{x(t=1)} y dx`, i.e. the area between the x
       axis and the curve, where :math:`y = y(t)` (the Bezier curve). By substitution
       :math:`dx = x'(t) dt`, performing the integration and
       subtracting the trapezoid we already considered above, we find (with control
       points :math:`(x_{c1}, y_{c1})` and :math:`(x_{c2}, y_{c2})`)

       .. math::

           \Delta A &= \int_0^1 y(t) x'(t) dt - \frac{1}{2} (y_1 + y_0) (x_1 - x_0) \\
           &= \frac{3}{20} \cdot \begin{pmatrix}
                  & y_0(& & 2x_{c1} & + x_{c2} & -3x_1&) \\
                + & y_{c1}(& -2x_0 & & + x_{c2} &+ x_1&) \\
                + & y_{c2}(& -x_0 & -x_{c1} & & + 2x_1&) \\
                + & y_1(& 3x_0 & - x_{c1} & -2 x_{c2} &&)
           \end{pmatrix}

       This is computed for every bezier and added to the area. Again, this is a signed
       area: convex beziers have a positive area and concave ones a negative area
       contribution.
    """

    MAT_AREA = numpy.array(
        [[0, 2, 1, -3], [-2, 0, 1, 1], [-1, -1, 0, 2], [3, -1, -2, 0]]
    )
    area = 0.0
    for sp in csp:
        if len(sp) < 2:
            continue
        for x, coord in enumerate(sp):  # calculate polygon area
            area += 0.5 * sp[x - 1][1][0] * (coord[1][1] - sp[x - 2][1][1])
        for i in range(1, len(sp)):  # add contribution from cubic Bezier
            # EXPLANATION: https://github.com/Pomax/BezierInfo-2/issues/238#issue-554619801
            vec_x = numpy.array(
                [sp[i - 1][1][0], sp[i - 1][2][0], sp[i][0][0], sp[i][1][0]]
            )
            vec_y = numpy.array(
                [sp[i - 1][1][1], sp[i - 1][2][1], sp[i][0][1], sp[i][1][1]]
            )
            vex = numpy.matmul(vec_x, MAT_AREA)
            area += 0.15 * numpy.matmul(vex, vec_y.T)
    return -area


def cspcofm(csp):
    r"""Get center of area / gravity for a cubic superpath.

    .. hint::

        The results may be slightly inaccurate for paths containing arcs due
        to the loss of accuracy during arc -> cubic bezier conversion.

    The function works similar to :func:`csparea`, only the computations are a bit more
    difficult. Again all subpaths are considered. The total center of mass is given by

    .. math::

        C_y = \frac{1}{A} \int_A y dA

    The integral can be expressed as a weighted sum; first, the contributions
    of the polygon created by the path's nodes is computed. Second, we compute the
    contribution of the Bezier curve; this is again done by an integral from which
    the weighted CofM of the trapezoid between end points and horizontal axis is
    removed. For the integrals, we have

    .. math::

        A * C_{y,bez} &= \int_A y dA = \int_{x(t=0)}^{y(t=1)} \int_{0}^{y(x)} y dy dx \\
        &= \int_{x(t=0)}^{y(t=1)} \frac 12 y(x)^2 dx
        = \int_0^1 \frac 12 y(t)^2  x'(t) dt \\
        A * C_{x,bez} &= \int_A x dA = \int_{x(t=0)}^{y(t=1)} x \int_{0}^{y(x)} dy dx \\
        &= \int_{x(t=0)}^{y(t=1)} x y(x) dx = \int_0^1 x(t) y(t) x'(t) dt

    from which the trapezoids are removed, in case of the y-CofM this amounts to

    .. math::

        \frac{y_0}{2} (x_1-x_0)y_0 + \left(y_0 + \frac 13 (y_1 - y_0)\right)
        \cdot  \frac 12 (y_1 - y_0) (x_1 - x_0)

    """

    MAT_COFM_0 = numpy.array(
        [[0, 35, 10, -45], [-35, 0, 12, 23], [-10, -12, 0, 22], [45, -23, -22, 0]]
    )

    MAT_COFM_1 = numpy.array(
        [[0, 15, 3, -18], [-15, 0, 9, 6], [-3, -9, 0, 12], [18, -6, -12, 0]]
    )

    MAT_COFM_2 = numpy.array(
        [[0, 12, 6, -18], [-12, 0, 9, 3], [-6, -9, 0, 15], [18, -3, -15, 0]]
    )

    MAT_COFM_3 = numpy.array(
        [[0, 22, 23, -45], [-22, 0, 12, 10], [-23, -12, 0, 35], [45, -10, -35, 0]]
    )
    area = csparea(csp)
    xc = 0.0
    yc = 0.0
    if abs(area) < 1.0e-8:
        raise ValueError(_("Area is zero, cannot calculate Center of Mass"))
    for sp in csp:
        for x, coord in enumerate(sp):  # calculate polygon moment
            xc += (
                sp[x - 1][1][1]
                * (sp[x - 2][1][0] - coord[1][0])
                * (sp[x - 2][1][0] + sp[x - 1][1][0] + coord[1][0])
                / 6
            )
            yc += (
                sp[x - 1][1][0]
                * (coord[1][1] - sp[x - 2][1][1])
                * (sp[x - 2][1][1] + sp[x - 1][1][1] + coord[1][1])
                / 6
            )
        for i in range(1, len(sp)):  # add contribution from cubic Bezier
            vec_x = numpy.array(
                [sp[i - 1][1][0], sp[i - 1][2][0], sp[i][0][0], sp[i][1][0]]
            )
            vec_y = numpy.array(
                [sp[i - 1][1][1], sp[i - 1][2][1], sp[i][0][1], sp[i][1][1]]
            )

            def _mul(MAT, vec_x=vec_x, vec_y=vec_y):
                return numpy.matmul(numpy.matmul(vec_x, MAT), vec_y.T)

            vec_t = numpy.array(
                [_mul(MAT_COFM_0), _mul(MAT_COFM_1), _mul(MAT_COFM_2), _mul(MAT_COFM_3)]
            )
            xc += numpy.matmul(vec_x, vec_t.T) / 280
            yc += numpy.matmul(vec_y, vec_t.T) / 280
    return -xc / area, -yc / area



#!/usr/bin/env python3
# coding=utf-8

import threading
import webbrowser

import inkex
from inkex import Anchor


class ThreadWebsite(threading.Thread):
    """Visit the website without locking inkscape"""

    def __init__(self, url):
        threading.Thread.__init__(self)
        self.url = url

    def run(self):
        webbrowser.open(self.url)


class FollowLink(inkex.EffectExtension):
    """Get the first selected item and follow it's href/url"""

    def effect(self):
        for node in self.svg.selection.filter(Anchor):
            vwswli = ThreadWebsite(node.get("xlink:href"))
            vwswli.start()
            break


if __name__ == "__main__":
    FollowLink().run()



#!/usr/bin/env python3
# coding=utf-8
#
# Copyright (c) 2024 jonathan.neuhauser@outlook.com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
Parse binary CGM files
"""

from dataclasses import dataclass, fields, is_dataclass
from io import BufferedIOBase, BytesIO
import struct
from typing import (
    ClassVar,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    get_type_hints,
    get_origin,
    get_args,
)
import warnings
import cgm_enums
import inkex


class FunctionRegistry(dict):
    def register(self, key):
        def decorator(func):
            self[key] = func
            return func

        return decorator


class BinaryCGMCommandParser:
    type_registry: Dict[Tuple[int, int], Type] = {}
    unsigned_readers = FunctionRegistry()
    signed_readers = FunctionRegistry()
    real_readers = FunctionRegistry()
    readers = FunctionRegistry()
    sizers = FunctionRegistry()

    def __init__(self, data: Union[bytes, BufferedIOBase]):
        if isinstance(data, bytes):
            data = BytesIO(data)
        # The stream we currently operate on. For long form commands its
        # easier to temporarily exchange this reference
        self.stream = data
        # The main stream of the file
        self.outer_stream = data
        # Reader functions for basic data types, where precision can be changed
        # in the file.
        self.integer_precision = 16
        self.index_precision = 16
        self.real_type = cgm_enums.RealTypeEnum.FIXED_POINT
        self.real_precision = (16, 16)
        self.vdc_integer_precision = 16
        self.vdc_real_type = cgm_enums.RealTypeEnum.FIXED_POINT
        self.vdc_real_precision = (16, 16)
        self.float_precision = (9, 23)
        self.colour_component_precision = 8
        self.colour_index_precision = 8
        self.vdc_type = cgm_enums.VDCTypeEnum.INTEGER
        self.vdc_real_type = cgm_enums.RealTypeEnum.FIXED_POINT
        self.vdc_real_precision = (16, 16)
        self.colour_model = cgm_enums.ColourModelEnum.RGB
        self.colour_selection_mode = cgm_enums.ColourSelectionModeEnum.INDEXED
        self.line_width_specification_mode = cgm_enums.WidthSpecificationModeEnum.SCALED
        self.marker_size_specification_mode = (
            cgm_enums.WidthSpecificationModeEnum.SCALED
        )
        self.edge_width_specification_mode = cgm_enums.WidthSpecificationModeEnum.SCALED
        self.interior_style_specification_mode = (
            cgm_enums.WidthSpecificationModeEnum.ABSOLUTE
        )

    @staticmethod
    def register_target_datatype(cls):
        if cls.id is not None:
            BinaryCGMCommandParser.type_registry[cls.id] = cls

    def read(self, bytes):
        res = self.stream.read(bytes)
        if len(res) != bytes:
            raise ValueError("EOF")
        return res

    def read_padded(self, length):
        # An uneven length of a command is padded with a null byte
        if length % 2 == 1:
            arg_bytes = self.read(length + 1)
            assert arg_bytes[-1] == 0
            return arg_bytes[:-1]
        return self.read(length)

    @unsigned_readers.register(8)
    def read_u8(self) -> int:
        return int.from_bytes(self.read(1), byteorder="big")

    @unsigned_readers.register(16)
    def read_u16(self) -> int:
        return int.from_bytes(self.read(2), byteorder="big")

    @unsigned_readers.register(24)
    def read_u24(self) -> int:
        return int.from_bytes(self.read(3), byteorder="big")

    @unsigned_readers.register(32)
    def read_u32(self) -> int:
        return int.from_bytes(self.read(4), byteorder="big")

    @signed_readers.register(8)
    def read_i8(self) -> int:
        return int.from_bytes(self.read(1), byteorder="big", signed=True)

    @signed_readers.register(16)
    @readers.register("E")
    def read_i16(self) -> int:
        return int.from_bytes(self.read(2), byteorder="big", signed=True)

    @signed_readers.register(24)
    def read_i24(self) -> int:
        return int.from_bytes(self.read(3), byteorder="big", signed=True)

    @signed_readers.register(32)
    def read_i32(self) -> int:
        return int.from_bytes(self.read(4), byteorder="big", signed=True)

    @real_readers.register((cgm_enums.RealTypeEnum.FLOATING_POINT, 9, 23))
    def read_float(self) -> float:
        return struct.unpack(">f", self.read(4))[0]

    @real_readers.register((cgm_enums.RealTypeEnum.FLOATING_POINT, 12, 52))
    def read_double(self) -> float:
        return struct.unpack(">d", self.read(8))[0]

    @real_readers.register((cgm_enums.RealTypeEnum.FIXED_POINT, 16, 16))
    def read_fixed_single(self) -> float:
        # Some sample files appear to not be using
        # two's complement for signed integers if they occur in a flot.
        # IMO that's incorrect according to the specification.
        return self.read_i16() + self.read_u16() / 2**16

    @real_readers.register((cgm_enums.RealTypeEnum.FIXED_POINT, 32, 32))
    def read_fixed_double(self) -> float:
        return self.read_i32() + self.read_u32() / 2**32

    @readers.register("S")
    @readers.register("D")
    @readers.register("SF")
    def read_str(self) -> str:
        """The first byte indicates the size.
        If the size is 255, the first two bytes indicate the size.
        From the new size, the first bit is a continuation flag
        and the remaining 15 bits are the size"""
        data: bytes
        size = self.read_u8()
        if size != 0xFF:
            data = self.read(size)
        else:
            # Long-form string
            self.stream.seek(0)
            _data = bytearray()
            while True:
                cur_size = self.read_u16()
                length = cur_size & 0b0111111111111111
                _data.extend(self.read(length))
                if not cur_size & 0b1000000000000000:
                    break
            data = bytes(_data)
        # TODO implement proper string encoding
        return data.decode("latin-1")

    def get_single_value(self, type_str: Union[str, Tuple[str]]):
        if type_str in self.readers:
            return BinaryCGMCommandParser.readers[type_str](self)
        else:
            raise ValueError(f"Unknown data type {type_str}")

    def get_value(
        self,
        type_str: Union[str, Tuple[str, ...]],
        target_types: Union[type, Tuple[type]],
        length: Optional[int] = None,
    ):
        """This function reads a value from the stream.
        The result can either be:
         - a primitive type (e.g. "I", <type int> -> int)
         - a tuple of primitive types
           (e.g. ("I", "P"), (<type int>, <type Point>) -> (int, Point))
         - a dataclass
           (e.g. ("P", "E"), <type PolygonLineSetEntry> -> PolygonLineSetEntry)
           in this case the type_str applies to the fields of the dataclass,
           and the target types of those fields are inferred from the
           type hint of the dataclass field
         - a list of any of the previous
           (e.g. ("I", "R"), typing.List[<type int>, <type float>]
           -> List[Tuple[int, float]])
           length must be defined for lists
           This just calls get_value with the argument of the List typehint,
           and collects the result until we've read length bytes
        """
        res = None
        initial_pos = self.stream.tell()

        if isinstance(type_str, str) and isinstance(target_types, type):
            if target_types == str and length == 0:
                return ""
            res = self.get_single_value(type_str)
            if type_str not in ("P", "CO", "CD", "CI"):
                res = target_types(res)

        elif isinstance(type_str, tuple):
            if get_origin(target_types) == tuple:
                target_types = get_args(target_types)
            if isinstance(target_types, tuple):
                assert len(target_types) == len(type_str)
                res = tuple(
                    self.get_value(i, t) for t, i in zip(target_types, type_str)
                )

            elif is_dataclass(target_types):
                res = target_types(
                    *self.get_value(
                        type_str,
                        tuple(
                            get_type_hints(target_types)[t.name]
                            for t in fields(target_types)
                        ),
                    )
                )
        if res is None and get_origin(target_types) == list:
            assert length is not None
            target_type = get_args(target_types)[0]
            try:
                field_size = self.get_size(type_str)
                # Assert that the remaining bytes fit into the list without residue
                assert length % field_size == 0
            except ValueError:
                # just an assertion here, continue
                pass
            res = []
            while True:
                if (self.stream.tell() - initial_pos) >= length:
                    break
                res.append(self.get_value(type_str, target_type))

        assert res is not None, "invalid specification for get_value"
        # assert isinstance(res, target_types), "bad return type"
        return res

    @readers.register("IX")
    def read_index(self):
        return BinaryCGMCommandParser.signed_readers[self.index_precision](self)

    @readers.register("SI")
    @readers.register("I")
    def read_signed(self):
        return BinaryCGMCommandParser.signed_readers[self.integer_precision](self)

    @readers.register("VDC")
    def read_vdc(self):
        return (
            BinaryCGMCommandParser.signed_readers[self.vdc_integer_precision](self)
            if self.vdc_type == cgm_enums.VDCTypeEnum.INTEGER
            else BinaryCGMCommandParser.real_readers[
                (self.vdc_real_type, *self.vdc_real_precision)
            ](self)
        )

    @readers.register("R")
    def read_real(self):
        return BinaryCGMCommandParser.real_readers[
            (self.real_type, *self.real_precision)
        ](self)

    @readers.register("F")
    def read_floating(self):
        return BinaryCGMCommandParser.real_readers[
            (cgm_enums.RealTypeEnum.FLOATING_POINT, *self.float_precision)
        ](self)

    @readers.register("CI")
    def read_indexed_colour(self):
        return cgm_enums.IndexColour(
            self.signed_readers[self.colour_index_precision](self)
        )

    @readers.register("CCO")
    def read_colour_component(self):
        return BinaryCGMCommandParser.unsigned_readers[self.colour_component_precision](
            self
        )

    @readers.register("CD")
    def read_direct_colour(self):
        result = cgm_enums.DirectColour(
            *(
                self.read_colour_component()
                for _ in range(
                    4 if self.colour_model == cgm_enums.ColourModelEnum.CMYK else 3
                )
            )
        )
        result.colour_model = self.colour_model
        return result

    @readers.register("CO")
    def read_colour(self):
        if self.colour_selection_mode == cgm_enums.ColourSelectionModeEnum.DIRECT:
            return self.read_direct_colour()
        else:
            return self.read_indexed_colour()

    @readers.register("P")
    def read_point(self):
        return cgm_enums.Point(self.read_vdc(), self.read_vdc())

    def read_size(self, mode: cgm_enums.WidthSpecificationModeEnum):
        if mode == cgm_enums.WidthSpecificationModeEnum.ABSOLUTE:
            return self.read_vdc()
        return self.read_real()

    @readers.register("LSS")
    def read_line_size(self):
        return self.read_size(self.line_width_specification_mode)

    @readers.register("MSS")
    def read_marker_size(self):
        return self.read_size(self.marker_size_specification_mode)

    @readers.register("ESS")
    def read_edge_size(self):
        return self.read_size(self.edge_width_specification_mode)

    @readers.register("ISS")
    def read_interior_style_size(self):
        return self.read_size(self.interior_style_specification_mode)

    # SIZING FUNCTIONS

    def get_size(self, type_str: Union[str, Tuple[str, ...]]):
        if type_str in BinaryCGMCommandParser.sizers:
            return BinaryCGMCommandParser.sizers[type_str](self)
        if isinstance(type_str, tuple):
            return sum(self.get_size(i) for i in type_str)
        else:
            raise ValueError(f"No size know for type {type_str}")

    @sizers.register("CI")
    def ci_size(self):
        return self.colour_index_precision // 8

    @sizers.register("CCO")
    def cco_size(self):
        return self.colour_component_precision // 8

    @sizers.register("CD")
    def cd_size(self):
        return self.cco_size() * (
            4 if self.colour_model == cgm_enums.ColourModelEnum.CMYK else 3
        )

    @sizers.register("IX")
    def ix_size(self):
        return self.index_precision // 8

    @sizers.register("E")
    def e_size(self):
        return 2

    @sizers.register("I")
    def i_size(self):
        return self.integer_precision // 8

    @sizers.register("R")
    def r_size(self):
        return (self.real_precision[0] + self.real_precision[1]) // 8

    @sizers.register("VDC")
    def vdc_size(self):
        return (
            self.vdc_integer_precision // 8
            if self.vdc_type == cgm_enums.VDCTypeEnum.INTEGER
            else (self.vdc_real_precision[0] + self.vdc_real_precision[1]) // 8
        )

    @sizers.register("P")
    def p_size(self):
        return 2 * self.vdc_size()

    @sizers.register("CO")
    def co_size(self):
        if self.colour_selection_mode == cgm_enums.ColourSelectionModeEnum.DIRECT:
            return self.cd_size()
        else:
            return self.ci_size()

    def ss_size(self, mode: cgm_enums.WidthSpecificationModeEnum):
        if mode == cgm_enums.WidthSpecificationModeEnum.ABSOLUTE:
            return self.vdc_size()
        return self.r_size()

    @sizers.register("LSS")
    def lss_size(self):
        return self.ss_size(self.line_width_specification_mode)

    @sizers.register("MSS")
    def marker_size(self):
        return self.ss_size(self.marker_size_specification_mode)

    @sizers.register("ESS")
    def edge_size(self):
        return self.ss_size(self.edge_width_specification_mode)

    @sizers.register("ISS")
    def interior_style_size(self):
        return self.ss_size(self.interior_style_specification_mode)

    def parse(self):
        while True:
            try:
                header = self.read_u16()
            except ValueError:
                return
            clss = header >> 12
            obj_id = (header & 0b0000111111100000) >> 5
            total_length = header & 0x1F

            if total_length == 31:
                # Long-form header
                args = bytearray()
                total_length = 0
                while True:
                    data = self.read_u16()
                    length = data & 0b0111111111111111
                    total_length += length
                    args.extend(self.read_padded(length))
                    if not data & 0b1000000000000000:
                        break
                self.stream = BytesIO(bytes(args))

            old_pos = self.stream.tell()
            interpreted = self.command_factory(clss, obj_id, total_length)
            new_pos = self.stream.tell()
            if (
                new_pos - old_pos not in [total_length] + [total_length - 1]
                if total_length % 2 == 0
                else []
            ):
                warnings.warn(
                    f"{interpreted.__class__.__name__}: Expected to read {total_length} bytes, but read {new_pos - old_pos}"
                )
            new_pos = old_pos + total_length
            self.stream.seek(new_pos)
            if isinstance(interpreted, ParserModifierCommand):
                self.update_readers(interpreted)

            # Prepare the stream for the next entry
            if self.stream == self.outer_stream:
                if new_pos % 2 != 0:
                    # Fields are always aligned with 2-byte words
                    self.read(1)
            else:
                self.stream = self.outer_stream

            yield interpreted

            if isinstance(interpreted, EndMetafile):
                # Sometimes there are trailing characters afterwards, return
                return

    def parse_data(self, data: BytesIO):
        """Temporarily exchanges the stream by data to read primitives from it,
        using the same settings as the current stream"""
        outer_stream = self.outer_stream
        stream = self.stream

        self.outer_stream = data
        self.stream = data
        result = tuple(self.parse())

        self.outer_stream = outer_stream
        self.stream = stream

        return result

    def update_readers(self, interpreted):
        if isinstance(interpreted, VdcType):
            self.vdc_type = interpreted.vdc_type
        elif isinstance(interpreted, IntegerPrecision):
            self.integer_precision = interpreted.integer_precision
        elif isinstance(interpreted, RealPrecision):
            self.real_type = interpreted.representation
            self.real_precision = (
                interpreted.exponent_whole_size,
                interpreted.fraction_size,
            )
            if interpreted.representation == cgm_enums.RealTypeEnum.FLOATING_POINT:
                self.float_precision = (
                    interpreted.exponent_whole_size,
                    interpreted.fraction_size,
                )
        elif isinstance(interpreted, IndexPrecision):
            self.index_precision = interpreted.index_precision
        elif isinstance(interpreted, ColourIndexPrecision):
            self.colour_index_precision = interpreted.colour_index_precision
        elif isinstance(interpreted, ColourPrecision):
            self.colour_component_precision = interpreted.colour_precision
        elif isinstance(interpreted, ColourSelectionMode):
            self.colour_selection_mode = interpreted.colour_selection_mode
        elif isinstance(interpreted, ColourModel):
            self.colour_model = interpreted.colour_model
        elif isinstance(interpreted, VDCIntegerPrecision):
            self.vdc_integer_precision = interpreted.integer_precision
        elif isinstance(interpreted, VDCRealPrecision):
            self.vdc_real_type = interpreted.representation
            self.vdc_real_precision = (
                interpreted.exponent_whole_size,
                interpreted.fraction_size,
            )
        elif isinstance(interpreted, LineWidthSpecificationMode):
            self.line_width_specification_mode = interpreted.size_specification
        elif isinstance(interpreted, EdgeWidthSpecificationMode):
            self.edge_width_specification_mode = interpreted.size_specification
        elif isinstance(interpreted, MarkerSizeSpecificationMode):
            self.marker_size_specification_mode = interpreted.size_specification
        elif isinstance(interpreted, InteriorStyleSpecificationMode):
            self.interior_style_specification_mode = interpreted.size_specification
        else:
            inkex.errormsg(
                f"Parser modifier command {interpreted} without associated action"
            )

    def command_factory(self, group, command, length):
        cls = self.type_registry.get((group, command), UnknownCommand)
        result = cls.parse_binary(self, length)
        if isinstance(result, UnknownCommand):
            result.id = (group, command)
        return result


@dataclass
class CGMCommand:
    id: ClassVar[Tuple[int, int]] = (0, 0)
    binary_params: ClassVar[Tuple[Union[str, Tuple[str, ...]], ...]] = tuple()

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        BinaryCGMCommandParser.register_target_datatype(cls)

    @classmethod
    def get_binary_args(
        cls,
        stream: BinaryCGMCommandParser,
        length: int,
    ):
        initial_pos = stream.stream.tell()
        for i, (field, types) in enumerate(zip(fields(cls), cls.binary_params)):
            outer_type = get_type_hints(cls)[field.name]
            yield stream.get_value(
                types, outer_type, length - (stream.stream.tell() - initial_pos)
            )

    @classmethod
    def parse_binary(cls, stream: BinaryCGMCommandParser, length):
        args = list(cls.get_binary_args(stream, length))
        return cls(*args)


6


@dataclass
class NoOp(CGMCommand):
    id = (0, 0)
    data: bytes

    @classmethod
    def parse_binary(cls, stream: BinaryCGMCommandParser, length):
        return cls(stream.read(length))


@dataclass
class UnknownCommand(CGMCommand):
    data: bytes

    @classmethod
    def parse_binary(cls, stream: BinaryCGMCommandParser, length):
        return cls(stream.read(length))


@dataclass
class BeginMetafile(CGMCommand):
    id = (0, 1)
    binary_params = ("SF",)
    name: str


@dataclass
class EndMetafile(CGMCommand):
    id = (0, 2)


@dataclass
class BeginPicture(CGMCommand):
    id = (0, 3)
    binary_params = ("SF",)
    name: str


@dataclass
class BeginPictureBody(CGMCommand):
    id = (0, 4)


@dataclass
class EndPicture(CGMCommand):
    id = (0, 5)


# Metafile descriptor elements


@dataclass
class MetafileVersion(CGMCommand):
    id = (1, 1)
    binary_params = ("I",)
    version: int


@dataclass
class MetafileDescription(CGMCommand):
    id = (1, 2)
    binary_params = ("SF",)
    description: str


class ParserModifierCommand(CGMCommand):
    pass


@dataclass
class VdcType(ParserModifierCommand):
    id = (1, 3)
    binary_params = ("E",)
    vdc_type: cgm_enums.VDCTypeEnum


@dataclass
class IntegerPrecision(ParserModifierCommand):
    id = (1, 4)
    binary_params = ("I",)
    integer_precision: int


@dataclass
class RealPrecision(ParserModifierCommand):
    id = (1, 5)
    binary_params = "E", "I", "I"
    representation: cgm_enums.RealTypeEnum
    exponent_whole_size: int
    fraction_size: int


@dataclass
class IndexPrecision(ParserModifierCommand):
    id = (1, 6)
    binary_params = ("I",)
    index_precision: int


@dataclass
class ColourPrecision(ParserModifierCommand):
    id = (1, 7)
    binary_params = ("I",)
    colour_precision: int


@dataclass
class ColourIndexPrecision(ParserModifierCommand):
    id = (1, 8)
    binary_params = ("I",)
    colour_index_precision: int


@dataclass
class MaximumColourIndex(CGMCommand):
    id = (1, 9)
    binary_params = ("CI",)
    maximum_colour_index: cgm_enums.IndexColour


@dataclass
class ColourValueExtent(CGMCommand):
    id = (1, 10)

    minimum: Optional[int]
    maximum: Optional[int]

    first_component: Optional[Tuple[float, float]]
    second_component: Optional[Tuple[float, float]]
    third_component: Optional[Tuple[float, float]]

    @classmethod
    def get_binary_args(cls, stream: BinaryCGMCommandParser, length):
        if stream.colour_model in (
            cgm_enums.ColourModelEnum.CMYK,
            cgm_enums.ColourModelEnum.RGB,
        ):
            return stream.read_direct_colour(), stream.read_direct_colour(), 0, 0, 0
        else:
            return 0, 0, stream.read_real(), stream.read_real(), stream.read_real()


@dataclass
class MetafileElementList(CGMCommand):
    id = (1, 11)
    elements: List[Tuple[int, int]]

    @classmethod
    def get_binary_args(cls, stream: BinaryCGMCommandParser, length: int):
        num = stream.read_signed()
        yield [(stream.read_index(), stream.read_index()) for _ in range(num)]


@dataclass
class MetafileDefaultsReplacement(CGMCommand):
    id = (1, 12)
    data: List[CGMCommand]

    @classmethod
    def get_binary_args(cls, stream: BinaryCGMCommandParser, length: int):
        data = stream.read(length)

        yield stream.parse_data(BytesIO(data))


@dataclass
class FontList(CGMCommand):
    id = (1, 13)
    binary_params = ("SF",)

    fonts: List[str]


@dataclass
class CharacterSetList(CGMCommand):
    id = (1, 14)
    binary_params = (("E", "SF"),)

    specification: List[Tuple[cgm_enums.CharacterSetTypeEnum, str]]


@dataclass
class CharacterCodingAnnouncer(CGMCommand):
    id = (1, 15)
    binary_params = ("E",)

    encoding: cgm_enums.CharacterCodingEnum


@dataclass
class MaximumVDCExtent(CGMCommand):
    id = (1, 17)
    binary_params = "P", "P"

    first_corner: cgm_enums.Point
    second_corner: cgm_enums.Point


@dataclass
class ColourModel(ParserModifierCommand):
    id = (1, 19)
    binary_params = ("IX",)

    colour_model: cgm_enums.ColourModelEnum


@dataclass
class ScalingMode(CGMCommand):
    id = (2, 1)
    binary_params = "E", "F"

    scaling_mode: cgm_enums.ScalingModeEnum
    metric_scaling_factor: float


@dataclass
class ColourSelectionMode(ParserModifierCommand):
    id = (2, 2)
    binary_params = ("E",)

    colour_selection_mode: cgm_enums.ColourSelectionModeEnum


@dataclass
class LineWidthSpecificationMode(ParserModifierCommand):
    id = (2, 3)
    binary_params = ("E",)

    size_specification: cgm_enums.WidthSpecificationModeEnum


@dataclass
class MarkerSizeSpecificationMode(ParserModifierCommand):
    id = (2, 4)

    binary_params = ("E",)
    size_specification: cgm_enums.WidthSpecificationModeEnum


@dataclass
class EdgeWidthSpecificationMode(ParserModifierCommand):
    id = (2, 5)

    binary_params = ("E",)
    size_specification: cgm_enums.WidthSpecificationModeEnum


@dataclass
class VDCExtent(CGMCommand):
    id = (2, 6)
    binary_params = "P", "P"

    first_corner: cgm_enums.Point
    second_corner: cgm_enums.Point


@dataclass
class BackgroundColour(CGMCommand):
    id = (2, 7)
    binary_params = ("CD",)

    colour: cgm_enums.DirectColour


@dataclass
class InteriorStyleSpecificationMode(ParserModifierCommand):
    id = (2, 16)

    binary_params = ("E",)
    size_specification: cgm_enums.WidthSpecificationModeEnum


@dataclass
class VDCIntegerPrecision(ParserModifierCommand):
    id = (3, 1)
    binary_params = ("I",)
    integer_precision: int


@dataclass
class VDCRealPrecision(ParserModifierCommand):
    id = (3, 2)
    binary_params = "E", "I", "I"
    representation: cgm_enums.RealTypeEnum
    exponent_whole_size: int
    fraction_size: int


@dataclass
class Transparency(CGMCommand):
    id = (3, 4)
    binary_params = ("E",)
    indicator: cgm_enums.TransparencyEnum


@dataclass
class ClipRectangle(CGMCommand):
    id = (3, 5)
    binary_params = ("P", "P")
    p1: cgm_enums.Point
    p2: cgm_enums.Point


@dataclass
class ClipIndicator(CGMCommand):
    id = (3, 6)
    binary_params = ("E",)
    clip: cgm_enums.ClipIndicatorEnum


@dataclass
class Polyline(CGMCommand):
    id = (4, 1)
    binary_params = ("P",)
    points: List[cgm_enums.Point]


@dataclass
class DisjointPolyline(CGMCommand):
    id = (4, 2)
    binary_params = ("P",)
    points: List[cgm_enums.Point]


@dataclass
class Polymarker(CGMCommand):
    id = (4, 3)
    binary_params = ("P",)
    points: List[cgm_enums.Point]


@dataclass
class Text(CGMCommand):
    id = (4, 4)
    binary_params = ("P", "E", "S")
    point: cgm_enums.Point
    flag: cgm_enums.TextFinalFlag
    string: str


@dataclass
class RestrictedText(CGMCommand):
    id = (4, 5)
    binary_params = ("VDC", "VDC", "P", "E", "S")
    width: float
    height: float
    point: cgm_enums.Point
    flag: cgm_enums.TextFinalFlag
    string: str


@dataclass
class AppendText(CGMCommand):
    id = (4, 6)
    binary_params = ("E", "S")
    flag: cgm_enums.TextFinalFlag
    string: str


@dataclass
class Polygon(CGMCommand):
    id = (4, 7)
    binary_params = ("P",)
    points: List[cgm_enums.Point]


@dataclass
class PolygonSet(CGMCommand):
    id = (4, 8)
    points: List[cgm_enums.PolygonSetEntry]
    binary_params = (("P", "E"),)


# TODO: Cell array (complicated parsing with sub-byte values)


@dataclass
class GeneralizedDrawingPrimitive(CGMCommand):
    id = (4, 10)
    identifier: int
    points: List[cgm_enums.Point]
    data: bytes

    @classmethod
    def get_binary_args(cls, stream: BinaryCGMCommandParser, length: int):
        yield stream.read_signed()
        num = stream.read_signed()
        points = [stream.read_point() for _ in range(num)]
        yield points
        read = stream.i_size() * 2 + num * stream.p_size()
        yield stream.read(length - read)


@dataclass
class Rectangle(CGMCommand):
    id = (4, 11)
    binary_params = ("P", "P")
    p1: cgm_enums.Point
    p2: cgm_enums.Point


@dataclass
class Circle(CGMCommand):
    id = (4, 12)
    binary_params = ("P", "VDC")
    center: cgm_enums.Point
    radius: float


@dataclass
class CircularArc3Point(CGMCommand):
    id = (4, 13)
    binary_params = ("P", "P", "P")
    p1: cgm_enums.Point
    p2: cgm_enums.Point
    p3: cgm_enums.Point


@dataclass
class CircularArc3PointClose(CircularArc3Point):
    id = (4, 14)
    binary_params = ("P", "P", "P", "E")  # type: ignore
    closure: cgm_enums.ArcClosureEnum


@dataclass
class CircularArcCentre(CGMCommand):
    id = (4, 15)
    binary_params = ("P", "VDC", "VDC", "VDC", "VDC", "VDC")  # type: ignore
    centre: cgm_enums.Point
    delta_x_start: float
    delta_y_start: float
    delta_x_end: float
    delta_y_end: float
    radius: float


@dataclass
class CircularArcCentreClose(CircularArcCentre):
    id = (4, 16)
    binary_params = ("P", "VDC", "VDC", "VDC", "VDC", "VDC", "E")  # type: ignore
    closure: cgm_enums.ArcClosureEnum


@dataclass
class Ellipse(CGMCommand):
    id = (4, 17)
    binary_params = ("P", "P", "P")
    center: cgm_enums.Point
    point_1: cgm_enums.Point
    point_2: cgm_enums.Point


@dataclass
class EllipticalArc(Ellipse):
    id = (4, 18)
    binary_params = ("P", "P", "P", "VDC", "VDC", "VDC", "VDC")  # type: ignore
    delta_x_start: float
    delta_y_start: float
    delta_x_end: float
    delta_y_end: float


@dataclass
class EllipticalArcClose(EllipticalArc):
    id = (4, 19)
    binary_params = ("P", "P", "P", "VDC", "VDC", "VDC", "VDC", "E")  # type: ignore
    closure: cgm_enums.ArcClosureEnum


@dataclass
class Polybezier(CGMCommand):
    id = (4, 26)
    binary_params = ("E", "P")
    continuous: cgm_enums.PolybezierContinuityEnum
    points: List[cgm_enums.Point]


@dataclass
class LineType(CGMCommand):
    id = (5, 2)
    binary_params = ("IX",)
    type: cgm_enums.LineTypeEnum


@dataclass
class LineWidth(CGMCommand):
    id = (5, 3)
    binary_params = ("LSS",)
    width: float


@dataclass
class LineColour(CGMCommand):
    id = (5, 4)
    binary_params = ("CO",)
    colour: cgm_enums.Colour


@dataclass
class MarkerType(CGMCommand):
    id = (5, 6)
    binary_params = ("E",)
    type: cgm_enums.MarkerTypeEnum


@dataclass
class MarkerSize(CGMCommand):
    id = (5, 7)
    binary_params = ("MSS",)
    size: float


@dataclass
class MarkerColour(CGMCommand):
    id = (5, 8)
    binary_params = ("CO",)
    colour: cgm_enums.Colour


@dataclass
class TextFontIndex(CGMCommand):
    id = (5, 10)
    binary_params = ("IX",)
    index: int


@dataclass
class TextPrecision(CGMCommand):
    id = (5, 11)
    binary_params = ("E",)
    text_precision: cgm_enums.TextPrecisionEnum


@dataclass
class CharacterExpansionFactor(CGMCommand):
    id = (5, 12)
    binary_params = ("R",)
    factor: float


@dataclass
class CharacterSpacing(CGMCommand):
    id = (5, 13)
    binary_params = ("R",)
    spacing: float


@dataclass
class TextColour(CGMCommand):
    id = (5, 14)
    binary_params = ("CO",)
    colour: cgm_enums.Colour


@dataclass
class CharacterHeight(CGMCommand):
    id = (5, 15)
    binary_params = ("VDC",)
    height: float


@dataclass
class CharacterOrientation(CGMCommand):
    id = (5, 16)
    binary_params = ("VDC", "VDC", "VDC", "VDC")
    x_character_up: float
    y_character_up: float
    x_character_base: float
    y_character_base: float


@dataclass
class TextPath(CGMCommand):
    id = (5, 17)
    binary_params = ("E",)
    path: cgm_enums.TextPathEnum


@dataclass
class TextAlignment(CGMCommand):
    id = (5, 18)
    binary_params = ("E", "E", "R", "R")

    horizontal_alignment: cgm_enums.TextHorizontalAlignmentEnum
    vertical_alignment: cgm_enums.TextVerticalAlignmentEnum
    continous_horizontal_alignment: float
    continous_vertical_alignment: float


@dataclass
class CharacterSetIndex(CGMCommand):
    id = (5, 19)
    binary_params = ("IX",)

    index: int


@dataclass
class AlternateCharacterSetIndex(CGMCommand):
    id = (5, 20)
    binary_params = ("IX",)

    index: int


@dataclass
class InteriorStyle(CGMCommand):
    id = (5, 22)
    binary_params = ("E",)

    interior_style: cgm_enums.InteriorStyleEnum


@dataclass
class FillColour(CGMCommand):
    id = (5, 23)
    binary_params = ("CO",)

    fill: cgm_enums.Colour


@dataclass
class HatchIndex(CGMCommand):
    id = (5, 24)
    binary_params = ("IX",)
    index: int


@dataclass
class PatternIndex(CGMCommand):
    id = (5, 25)
    binary_params = ("IX",)
    index: int


@dataclass
class EdgeType(CGMCommand):
    id = (5, 27)
    binary_params = ("IX",)
    type: cgm_enums.LineTypeEnum


@dataclass
class EdgeWidth(CGMCommand):
    id = (5, 28)
    binary_params = ("ESS",)
    width: float


@dataclass
class EdgeColour(CGMCommand):
    id = (5, 29)
    binary_params = ("CO",)
    colour: cgm_enums.Colour


@dataclass
class EdgeVisibility(CGMCommand):
    id = (5, 30)
    binary_params = ("E",)
    visibility: cgm_enums.EdgeVisibilityEnum


@dataclass
class ColourTable(CGMCommand):
    id = (5, 34)
    binary_params = ("CI", "CD")

    colour_index: cgm_enums.IndexColour
    colours: List[cgm_enums.DirectColour]


@dataclass
class AspectSourceFlags(CGMCommand):
    id = (5, 35)
    binary_params = (("E", "E"),)
    flags: List[Tuple[cgm_enums.ASFTypeEnum, cgm_enums.ASFValue]]


@dataclass
class LineCap(CGMCommand):
    id = (5, 37)
    binary_params = ("E", "E")
    cap: cgm_enums.CapIndicatorEnum
    dash_cap: cgm_enums.DashCapIndicatorEnum


@dataclass
class RestrictedTextType(CGMCommand):
    id = (5, 42)
    binary_params = ("E",)
    type: cgm_enums.RestrictionTypeEnum


@dataclass
class ApplicationData(CGMCommand):
    id = (7, 2)
    binary_params = ("I", "SF")

    identifier: int
    data: str



#!/usr/bin/env python3
"""Reverse the colors"""

import inkex


class Negative(inkex.ColorExtension):
    """Make the colour oposite"""

    def modify_color(self, name, color):
        # Support any colour space
        for i, channel in enumerate(color):
            color[i] = 255 - channel
        return color


if __name__ == "__main__":
    Negative().run()



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024 Manpreet Singh <manpreet.singh.dev@proton.me>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from io import BytesIO
import math
import os
from typing import Optional, Union

from PIL import Image
import numpy as np

from inkaf.parser.types import AFObject
from inkaf.parser.extract import AFExtractor
from inkaf.utils import extract


@dataclass
class AFBitmap:
    format: int
    width: int
    height: int
    emb_file: Optional[Union[str, AFObject]]

    @classmethod
    def from_af(cls, child: AFObject) -> AFBitmap:
        if "Bckg" in child:
            bckg = child.get("Bckg")
        else:
            bckg = None

        # TODO:
        # icc_profiles: dict = child["Prof"]
        # rgb_profile: bytes = icc_profiles["RGBP"]
        # rgu_profile: bytes = icc_profiles["RGUP"]
        # cmyk_profile: bytes = icc_profiles["CMYP"]
        # ia_profile: bytes = icc_profiles["IAPr"]
        # lab_profile: bytes = icc_profiles["LABP"]
        # dela=child["DelA"],  # TODO: What is this?

        return AFBitmap(
            format=child["Frmt"].id,
            emb_file=bckg.data if bckg else child,
            width=child["BmpW"],
            height=child["BmpH"],
        )

    @classmethod
    def bitmap_from_blocks(cls, child: AFObject, extractor: AFExtractor):
        """Assemble a bitmap from 256² blocks"""
        size = (child["BmpH"], child["BmpW"])
        components = []
        for component in range(1, 5):
            if f"TWi{component}" in child:
                block_layout = (child[f"THi{component}"], child[f"TWi{component}"])
            else:
                block_layout = (math.ceil(size[0] / 256), math.ceil(size[1] / 256))

            data = np.zeros(
                (256 * block_layout[0], 256 * block_layout[1]), dtype=np.uint8
            )
            block_types = child[f"Sta{component}"]
            j = 0
            for i, block in enumerate(block_types):
                col = i % block_layout[1]
                row = i // block_layout[1]
                if block == 0:
                    blkdata = np.zeros((256, 256), dtype=np.uint8)
                elif block == 1:
                    blkdata = np.zeros((256, 256), dtype=np.uint8)
                elif block == 2:
                    blkdata = (np.ones((256, 256), dtype=np.uint8) * 255).astype(
                        np.uint8
                    )
                elif block == 4:
                    block = child[f"Idx{component}"][j]
                    assert block.get_type() == "Blck"
                    blkdata = np.frombuffer(
                        extract(extractor, block["Data"].data), dtype=np.uint8
                    ).reshape(256, 256)
                    j += 1
                else:
                    raise RuntimeError(f"Unknown block type {block}")
                data[row * 256 : (row + 1) * 256, col * 256 : (col + 1) * 256] = blkdata
            assert j == len(child[f"Idx{component}"])
            components.append(data[0 : size[0], 0 : size[1]])
        imagefile = BytesIO()
        image = Image.fromarray(np.stack(components, axis=-1), "RGBA")
        image.save(imagefile, format="PNG")
        imagefile.seek(0)
        return imagefile.read()


class AFImageTypes(Enum):
    JPEG = 0
    PNG = 1
    TIFF = 2
    WebP = 3
    HD_Photo = 4
    GIF = 6
    JPEG_2000 = 7
    BMP = 8
    OpenEXR = 10
    HDR = 11  # https://en.wikipedia.org/wiki/RGBE_image_format
    HEIF = 14
    JXL = 16  # file_type_str = plugin


media_types = {
    "JPEG": "image/jpeg",
    "PNG": "image/png",
    "TIFF": "image/tiff",
    "WebP": "image/webp",
    "HD_Photo": "image/jxr",
    "GIF": "image/gif",
    "JPEG_2000": "image/jp2",
    "BMP": "image/bmp",
    "HEIF": "image/heif",
    "JXL": "image/jxl",
}


@dataclass
class AFImage:
    bitmap: AFBitmap

    file_path_abs: Optional[str]
    file_path_rel: Optional[str]
    file_size: int
    file_type_enum: Optional[AFImageTypes]
    file_type_str: Optional[str]

    @property
    def media_type(self) -> Optional[str]:
        if self.file_type_enum is not None and self.file_type_enum.name in media_types:
            return media_types[self.file_type_enum.name]

        if self.file_type_str in media_types:
            return media_types[self.file_type_str]

        path = self.file_path_rel or self.file_path_abs
        if path:
            file_ext = os.path.splitext(path)[1][1:].lower()
            for k, v in media_types.items():
                if file_ext == k.lower() or file_ext == v.split("/")[1]:
                    return v
        return None

    @classmethod
    def from_af(cls, child: AFObject) -> AFImage:
        assert child.get_type() in ("ImgN", "Rstr"), "Not an image"
        # modification_time=child["MfTm"],
        # file_name=child["IRFN"],
        return cls(
            AFBitmap.from_af(child["Bitm"]),
            file_path_abs=(child.get("IRDS") or {}).get("Filn"),
            file_path_rel=(child.get("RPth") or {}).get("Path"),
            file_size=child.get("Flsz"),
            file_type_enum=AFImageTypes(child["FTyp"].id)
            if "FTyp" in child and child["FTyp"].id in [t.value for t in AFImageTypes]
            else None,
            file_type_str=child.get("FTyN"),
        )



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2024 Manpreet Singh <manpreet.singh.dev@proton.me>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from __future__ import annotations

import itertools
from dataclasses import dataclass
from enum import Enum
from typing import List, Optional, Tuple

import inkex

from ..parser.types import AFObject
from .curve import AFCurve
from .fill import AFColor, AFGradient, parse_bfil, parse_fdsc


class AFBlendModeV0(Enum):
    NORMAL = 0
    DARKEN = 1
    MULTIPLY = 2
    COLOR_BURN = 3
    LIGHTEN = 4
    SCREEN = 5
    COLOR_DODGE = 6
    ADD = 7
    OVERLAY = 8
    SOFT_LIGHT = 9
    HARD_LIGHT = 10
    VIVID_LIGHT = 11
    PIN_LIGHT = 12
    HARD_MIX = 13
    DIFFERENCE = 14
    EXCLUSION = 15
    SUBTRACT = 16
    HUE = 17
    SATURATION = 18
    LUMINOSITY = 19
    COLOR = 20
    AVERAGE = 21
    NEGATION = 22
    REFLECT = 23
    GLOW = 24
    ERASE = 25


class AFBlendModeV1(Enum):
    DARKER_COLOR = 2
    LIGHTER_COLOR = 6
    LINEAR_LIGHT = 15


class AFBlendModeV2(Enum):
    CONTRAST_NEGATE = 28


class AFBlendModeV3(Enum):
    LINEAR_BURN = 5


class AFBlendModeV4(Enum):
    DIVIDE = 21


AF_BLEND_MODES = {
    0: AFBlendModeV0,
    1: AFBlendModeV1,
    2: AFBlendModeV2,
    3: AFBlendModeV3,
    4: AFBlendModeV4,
}

SVG_BLEND_MODES = {
    "normal",
    "multiply",
    "screen",
    "overlay",
    "darken",
    "lighten",
    "color-dodge",
    "color-burn",
    "hard-light",
    "soft-light",
    "difference",
    "exclusion",
    "hue",
    "saturation",
    "color",
    "luminosity",
    "plus-darker",
    "plus-lighter",
}


class AFStrokeType(Enum):
    NONE = 0
    SOLID = 1
    DASHED = 2
    TEXTURED = 3


class AFStrokeCap(Enum):
    BUTT = 0
    SQUARE = 1
    ROUND = 2


class AFStrokeJoin(Enum):
    NONE = 0  # Unknown
    MITER = 1
    ROUND = 2
    BEVEL = 3


class AFStrokeAlign(Enum):
    CENTER = 0
    INSIDE = 1
    OUTSIDE = 2


class AFStrokeOrder(Enum):
    FRONT = False
    BEHIND = True


@dataclass
class AFStroke:
    width: float
    style: AFStrokeType
    cap: AFStrokeCap
    join: AFStrokeJoin
    order: AFStrokeOrder
    align: AFStrokeAlign
    miter_clip: float
    scale_with_object: bool
    dash_pattern: List[float]
    width_curve: Optional[AFCurve] = None

    def __post_init__(self):
        if self.width_curve and not self.width_curve.curves:
            self.width_curve = None

        # Width remains constant for DASHED strokes
        if self.width_curve and self.style not in [
            AFStrokeType.SOLID,
            AFStrokeType.TEXTURED,
        ]:
            multiplier = max(
                self.width_curve.curves[1], key=lambda node: node.point[1]
            ).point[1]
            self.width *= multiplier
            self.width_curve = None

    @classmethod
    def from_af(cls, liln: AFObject):
        assert liln.get_type() == "LDsc"
        return cls(
            width=float(liln["LDeL"]["Wght"]),
            cap=AFStrokeCap(liln["LDeL"]["Data"].unsigned_ints[0]),
            join=AFStrokeJoin(liln["LDeL"]["Data"].unsigned_ints[1]),
            style=AFStrokeType(liln["LDeL"]["Data"].unsigned_ints[2]),
            miter_clip=float(liln["LDeL"]["Data"].doubles),
            dash_pattern=liln["LDeL"].get("Patn", [1, 1, 0, 0, 0, 0]),
            order=AFStrokeOrder(liln["LDBe"]),
            align=AFStrokeAlign(liln["LDSa"]),
            scale_with_object=liln["LDSc"],
            width_curve=(
                AFCurve.from_af(liln["LDeP"]) if "Data" in liln["LDeP"] else None
            ),
            # TODO
            # liln["LIAh"],
            # liln["LDSc"],
            # liln["LDeL"]["Brus"],
            # liln["LDeL"]["DBal"],
        )

    def get_power_stroke(self, svg: inkex.ShapeElement) -> Optional[inkex.PathEffect]:
        if self.width_curve is None:
            return None
        path = svg.path
        locations = [i.point[0] for i in self.width_curve.curves[0]]
        # Get local lengths
        subpaths = path.break_apart()
        start_indices = [0] + list(itertools.accumulate(len(s) for s in subpaths))
        moves = 0
        resulting_offsets: List[Tuple[float, float]] = []
        for start_index, subpath in zip(start_indices, subpaths):
            # the same width curve is applied to all subpaths
            lengths = [
                command.length() if command.letter not in "Mm" else 0
                for command in subpath.proxy_iterator()
            ]
            cumulative_lengths = list(itertools.accumulate(lengths))
            total_length = cumulative_lengths[-1]
            result = {}
            moves += 1
            for (_, l1), ((i, p2), l2) in inkex.utils.pairwise(
                zip(enumerate(subpath.proxy_iterator()), cumulative_lengths),
                start=False,
            ):
                if l2 - l1 == 0:
                    moves += 1
                for location in locations:
                    if l1 <= location * total_length <= l2 and location not in result:
                        # On the current segment and not processed yet
                        # Coordinates are specified as
                        # no. visible segment + local path coordinate
                        result[location] = (start_index + i - moves) + p2.ilength(
                            location * total_length - l1
                        )
            resulting_offsets.extend(
                (result[point.point[0]], abs(point.point[1]))
                for point in self.width_curve.curves[0]
                if point.index == 0
            )
        width = svg.to_dimensionless(svg.style("stroke-width"))

        offset_pts = " | ".join(
            f"{location:.6f},{offset * width / 2:.6f}"
            for location, offset in resulting_offsets
        )

        assert len(self.width_curve.curves) == 1, "Invalid width curve"

        svg.set("inkscape:original-d", str(svg.path))
        return inkex.PathEffect.new(
            effect="powerstroke",
            is_visible="true",
            lpeversion="1.3",
            scale_width="1",
            interpolator_type="CubicBezierSmooth",
            interpolator_beta="0.22",
            start_linecap_type=stroke_cap_map[self.cap],
            end_linecap_type=stroke_cap_map[self.cap],
            linejoin_type=stroke_join_map[self.join],
            miter_limit=self.miter_clip,
            sort_points="true",
            not_jump="false",
            offset_points=offset_pts,
        )


stroke_cap_map = {e: e.name.lower() for e in AFStrokeCap}
stroke_join_map = {
    **{e: e.name.lower() for e in AFStrokeJoin},
    AFStrokeJoin.NONE: "miter",
}
stroke_order_map = {
    AFStrokeOrder.FRONT: "fill stroke markers",
    AFStrokeOrder.BEHIND: "stroke markers fill",
}


def parse_transform(xfrm: List[float]) -> inkex.Transform:
    return inkex.Transform((xfrm[:3], xfrm[3:]))


def parse_fill(child: AFObject) -> Optional[AFColor | AFGradient]:
    if "BFil" in child:
        return parse_bfil(child["BFil"])
    elif "BFCr" in child and "BFFl" in child:
        fill_idx = child["BFCr"]
        return parse_fdsc(child["BFFl"][fill_idx])
    else:
        return None


def parse_stroke_color(child: AFObject) -> Optional[AFColor | AFGradient]:
    if "LICr" in child and "LIFl" in child:
        stroke_idx = child["LICr"]
        return parse_fdsc(child["LIFl"][stroke_idx])
    else:
        return None


def parse_stroke(child) -> Optional[AFStroke]:
    # Stroke
    # TODO: Align, Textured style, Start-End markers
    if "LSty" in child:
        return AFStroke.from_af(child["LSty"])
    elif "LILn" in child and "LICr" in child:
        stroke_idx = child["LICr"]
        return AFStroke.from_af(child["LILn"][stroke_idx])
    else:
        return None
